# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

INSTANCE_GROUPS_WITH_AUTOSCALING_POLICY_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=TASK,InstanceType=d2.xlarge,InstanceCount=2,'
    'AutoScalingPolicy={Constraints={MinCapacity=1,MaxCapacity=2},'
    'Rules=[{Name=TestRule,Description=TestDescription,Action={Market=ON_DEMAND,'
    'SimpleScalingPolicyConfiguration={AdjustmentType=EXACT_CAPACITY,ScalingAdjustment=2,CoolDown=5}},'
    'Trigger={CloudWatchAlarmDefinition={ComparisonOperator=GREATER_THAN,'
    'EvaluationPeriods=5,MetricName=TestMetric,Namespace=EMR,Period=3,Statistic=MAXIMUM,'
    'Threshold=4.565,Unit=NONE,Dimensions=[{Key=TestKey,Value=TestValue}]}}}]}'
)

INSTANCE_GROUPS_WITH_EBS_VOLUME_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,SizeInGB=100,Iops=100},VolumesPerInstance=4},{VolumeSpecification={VolumeType=gp3,SizeInGB=100,Iops=3000,Throughput=200},VolumesPerInstance=1},{VolumeSpecification={VolumeType=gp3,SizeInGB=100}}]}')

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_VOLTYPE_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={SizeInGB=100,Iops=100},VolumesPerInstance=4},{VolumeSpecification={VolumeType=gp2,SizeInGB=100,Iops=100}}]}')

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_SIZE_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,Iops=100},VolumesPerInstance=4},{VolumeSpecification={VolumeType=gp2,SizeInGB=100,Iops=100}}]}')

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_VOLSPEC_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true}')

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_IOPS_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,SizeInGB=100},VolumesPerInstance=4}]}')

MULTIPLE_INSTANCE_GROUPS_WITH_EBS_VOLUMES_VOLUME_ARG = (
    ' InstanceGroupType=MASTER,InstanceCount=1,InstanceType=d2.xlarge,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={Iops=100,VolumeType=gp2,SizeInGB=20},VolumesPerInstance=4}]} '
    'InstanceGroupType=CORE,InstanceType=d2.xlarge,InstanceCount=2,EbsConfiguration={EbsOptimized=true,EbsBlockDeviceConfigs=[{VolumeSpecification={VolumeType=gp2,SizeInGB=100}}]}')

INSTANCE_GROUPS_WITH_EBS = \
    [{'InstanceCount': 1,
      'InstanceRole': 'MASTER',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'MASTER'},
     {'EbsConfiguration': {'EbsOptimized': True,
                           'EbsBlockDeviceConfigs': [{'VolumeSpecification':
                                                      {'Iops': 100,
                                                       'SizeInGB': 100,
                                                       'VolumeType': 'gp2'},
                                                      'VolumesPerInstance': 4
                                                     },
                                                     {'VolumeSpecification':
                                                       {'Iops': 3000,
                                                        'SizeInGB': 100,
                                                        'VolumeType': 'gp3',
                                                        'Throughput': 200 },
                                                      'VolumesPerInstance': 1
                                                     },
                                                     {'VolumeSpecification':
                                                       {'SizeInGB': 100,
                                                        'VolumeType': 'gp3'}}]},
      'InstanceCount': 2,
      'InstanceRole': 'CORE',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'CORE'}]

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_IOPS = \
    [{'InstanceCount': 1,
      'InstanceRole': 'MASTER',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'MASTER'},
     {'EbsConfiguration': {'EbsOptimized': True,
                           'EbsBlockDeviceConfigs': [{'VolumeSpecification':
                                                      {'SizeInGB': 100,
                                                       'VolumeType': 'gp2'},
                                                      'VolumesPerInstance': 4}]},
      'InstanceCount': 2,
      'InstanceRole': 'CORE',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'CORE'}]

INSTANCE_GROUPS_WITH_EBS_VOLUME_MISSING_VOLSPEC = \
    [{'InstanceCount': 1,
      'InstanceRole': 'MASTER',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'MASTER'},
     {'EbsConfiguration': {'EbsOptimized': True},
      'InstanceCount': 2,
      'InstanceRole': 'CORE',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'CORE'}]

MULTIPLE_INSTANCE_GROUPS_WITH_EBS_VOLUMES = \
    [{'EbsConfiguration': {'EbsOptimized': True,
                           'EbsBlockDeviceConfigs': [{'VolumeSpecification':
                                                      {'Iops': 100,
                                                       'SizeInGB': 20,
                                                       'VolumeType': 'gp2'},
                                                     'VolumesPerInstance': 4}]},
      'InstanceCount': 1,
      'InstanceRole': 'MASTER',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'MASTER'},
     {'EbsConfiguration': {'EbsOptimized': True,
                           'EbsBlockDeviceConfigs': [{'VolumeSpecification':
                                                      {'SizeInGB': 100,
                                                       'VolumeType': 'gp2'}}]},
      'InstanceCount': 2,
      'InstanceRole': 'CORE',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'CORE'}]

INSTANCE_GROUPS_WITH_AUTOSCALING_POLICY = \
    [{'InstanceCount': 1,
      'InstanceRole': 'MASTER',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'MASTER'},
     {'InstanceCount': 2,
      'InstanceRole': 'TASK',
      'InstanceType': 'd2.xlarge',
      'Market': 'ON_DEMAND',
      'Name': 'TASK',
      'AutoScalingPolicy': {
          'Constraints': {
              'MinCapacity': 1,
              'MaxCapacity': 2
          },
          'Rules': [
              {
                  'Name': 'TestRule',
                  'Description': 'TestDescription',
                  'Action': {
                      'Market': 'ON_DEMAND',
                      'SimpleScalingPolicyConfiguration': {
                          'AdjustmentType': 'EXACT_CAPACITY',
                          'ScalingAdjustment': 2,
                          'CoolDown': 5
                      }
                  },
                  'Trigger': {
                      'CloudWatchAlarmDefinition': {
                          'ComparisonOperator': 'GREATER_THAN',
                          'Dimensions': [
                              {
                                  'Key': 'TestKey',
                                  'Value': 'TestValue'
                              }],
                          'EvaluationPeriods': 5,
                          'MetricName': 'TestMetric',
                          'Namespace': 'EMR',
                          'Period': 3,
                          'Statistic': 'MAXIMUM',
                          'Threshold': 4.565,
                          'Unit': 'NONE'
                      }
                  }
              }
          ]
      }
      }]


# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import pytest
from tests import unittest

from botocore import regions
from botocore.exceptions import NoRegionError, EndpointVariantError


class TestEndpointResolver(unittest.TestCase):
    def _template(self):
        return {
            'partitions': [
                {
                    'partition': 'aws',
                    'dnsSuffix': 'amazonaws.com',
                    'regionRegex': r'^(us|eu)\-\w+$',
                    'defaults': {
                        'hostname': '{service}.{region}.{dnsSuffix}',
                        'variants': [
                            {
                                'hostname': '{service}-fips.{region}.{dnsSuffix}',
                                'tags': ['fips']
                            },
                            {
                                'dnsSuffix': 'api.aws',
                                'hostname': '{service}.{region}.{dnsSuffix}',
                                'tags': ['dualstack']
                            },
                            {
                                'dnsSuffix': 'api.aws',
                                'hostname': '{service}-fips.{region}.{dnsSuffix}',
                                'tags': ['dualstack', 'fips']
                            },
                        ]
                    },
                    'regions': {
                        'us-foo': {'regionName': 'a'},
                        'us-bar': {'regionName': 'b'},
                        'eu-baz': {'regionName': 'd'}
                    },
                    'services': {
                        'ec2': {
                            'defaults': {
                                'protocols': ['http', 'https'],
                                'variants': [
                                    {
                                        'dnsSuffix': 'api.aws',
                                        'hostname': 'api.ec2.{region}.{dnsSuffix}',
                                        'tags': ['dualstack']
                                    }
                                ]
                            },
                            'endpoints': {
                                'us-foo': {
                                    'hostname': 'ec2.us-foo.amazonaws.com',
                                    'variants': [
                                        {
                                            'dnsSuffix': 'api.aws',
                                            'hostname': 'ec2.foo.{dnsSuffix}',
                                            'tags': ['dualstack']
                                        },
                                        {
                                            'hostname': 'ec2-fips.foo.amazonaws.com',
                                            'tags': ['fips']
                                        },
                                        {
                                            'hostname': 'ec2-fips.foo.api.aws',
                                            'tags': ['fips', 'dualstack']
                                        },
                                    ]
                                },
                                'us-bar': {},
                                'us-dep': {
                                    'deprecated': True,
                                },
                                'us-fizz': {
                                    'credentialScope': {
                                        'region': 'us-fizz'
                                    },
                                    'hostname': 'ec2.us-fizz.amazonaws.com',
                                    'variants': [
                                        {
                                            'hostname': 'ec2.fizz.api.aws',
                                            'tags': ['dualstack']
                                        }
                                    ]
                                },
                                'eu-baz': {},
                                'd': {}
                            },
                        },
                        's3': {
                            'defaults': {
                                'sslCommonName': '{service}.{region}.{dnsSuffix}',
                                'variants': [
                                    {
                                        'hostname': 's3.dualstack.{region}.{dnsSuffix}',
                                        'tags': ['dualstack']
                                    },
                                    {
                                        'hostname': 's3-fips.{region}.{dnsSuffix}',
                                        'tags': ['fips']
                                    },
                                    {
                                        'hostname': 's3-fips.dualstack.{region}.{dnsSuffix}',
                                        'tags': ['dualstack', 'fips']
                                    }
                                ]
                            },
                            'endpoints': {
                                'us-foo': {
                                    'sslCommonName': \
                                        '{region}.{service}.{dnsSuffix}'
                                },
                                'us-bar': {},
                                'us-fizz': {
                                    'hostname': 's3.api.us-fizz.amazonaws.com',
                                    'variants': [
                                        {
                                            'tags': ['dualstack']
                                        },
                                        {
                                            'tags': ['fips']
                                        }
                                    ]
                                },
                                'eu-baz': {'hostname': 'foo'}
                            }
                        },
                        'not-regionalized': {
                            'isRegionalized': False,
                            'partitionEndpoint': 'aws',
                            'endpoints': {
                                'aws': {'hostname': 'not-regionalized'},
                                'us-foo': {},
                                'eu-baz': {}
                            }
                        },
                        'non-partition': {
                            'partitionEndpoint': 'aws',
                            'endpoints': {
                                'aws': {'hostname': 'host'},
                                'us-foo': {}
                            }
                        },
                        'merge': {
                            'defaults': {
                                'signatureVersions': ['v2'],
                                'protocols': ['http']
                            },
                            'endpoints': {
                                'us-foo': {'signatureVersions': ['v4']},
                                'us-bar': {'protocols': ['https']}
                            }
                        }
                    }
                },
                {
                    'partition': 'foo',
                    'dnsSuffix': 'foo.com',
                    'regionRegex': r'^(foo)\-\w+$',
                    'defaults': {
                        'hostname': '{service}.{region}.{dnsSuffix}',
                        'protocols': ['http'],
                        'foo': 'bar',
                    },
                    'regions': {
                        'foo-1': {'regionName': '1'},
                        'foo-2': {'regionName': '2'},
                        'foo-3': {'regionName': '3'}
                    },
                    'services': {
                        'ec2': {
                            'endpoints': {
                                'foo-1': {
                                    'foo': 'baz'
                                },
                                'foo-2': {},
                                'foo-3': {}
                            }
                        }
                    }
                },
                {
                    'partition': 'aws-iso',
                    'dnsSuffix': 'amazonaws.com',
                    'defaults': {
                        'hostname': '{service}.{region}.{dnsSuffix}',
                        'protocols': ['http'],
                    },
                    'regions': {
                        'foo-1': {'regionName': '1'},
                        'foo-2': {'regionName': '2'},
                        'foo-3': {'regionName': '3'}
                    },
                    'services': {
                        'ec2': {
                            'endpoints': {
                                'foo-1': {
                                    'foo': 'baz'
                                },
                                'foo-2': {},
                                'foo-3': {}
                            }
                        }
                    }
                }
            ]
        }

    def test_ensures_region_is_not_none(self):
        with self.assertRaises(NoRegionError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint('foo', None)

    def test_ensures_required_keys_present(self):
        with self.assertRaises(ValueError):
            regions.EndpointResolver({})

    def test_returns_empty_list_when_listing_for_different_partition(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual([], resolver.get_available_endpoints('ec2', 'bar'))

    def test_returns_empty_list_when_no_service_found(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual([], resolver.get_available_endpoints('what?'))

    def test_gets_endpoint_names(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.get_available_endpoints(
            'ec2', allow_non_regional=True)
        self.assertEqual(['d', 'eu-baz', 'us-bar', 'us-dep',
                          'us-fizz', 'us-foo'],
                         sorted(result))

    def test_gets_endpoint_names_for_partition(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.get_available_endpoints(
            'ec2', allow_non_regional=True, partition_name='foo')
        self.assertEqual(['foo-1', 'foo-2', 'foo-3'], sorted(result))

    def test_list_regional_endpoints_only(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.get_available_endpoints(
            'ec2', allow_non_regional=False)
        self.assertEqual(['eu-baz', 'us-bar', 'us-foo'], sorted(result))

    def test_returns_none_when_no_match(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertIsNone(resolver.construct_endpoint('foo', 'baz'))

    def test_constructs_regionalized_endpoints_for_exact_matches(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized', 'eu-baz')
        self.assertEqual('not-regionalized.eu-baz.amazonaws.com',
                          result['hostname'])
        self.assertEqual('aws', result['partition'])
        self.assertEqual('eu-baz', result['endpointName'])

    def test_constructs_partition_endpoints_for_real_partition_region(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized', 'us-bar')
        self.assertEqual('not-regionalized', result['hostname'])
        self.assertEqual('aws', result['partition'])
        self.assertEqual('aws', result['endpointName'])

    def test_constructs_partition_endpoints_for_regex_match(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized', 'us-abc')
        self.assertEqual('not-regionalized', result['hostname'])

    def test_constructs_endpoints_for_regionalized_regex_match(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'us-abc')
        self.assertEqual('s3.us-abc.amazonaws.com', result['hostname'])

    def test_constructs_endpoints_for_unknown_service_but_known_region(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('unknown', 'us-foo')
        self.assertEqual('unknown.us-foo.amazonaws.com', result['hostname'])

    def test_merges_service_keys(self):
        resolver = regions.EndpointResolver(self._template())
        us_foo = resolver.construct_endpoint('merge', 'us-foo')
        us_bar = resolver.construct_endpoint('merge', 'us-bar')
        self.assertEqual(['http'], us_foo['protocols'])
        self.assertEqual(['v4'], us_foo['signatureVersions'])
        self.assertEqual(['https'], us_bar['protocols'])
        self.assertEqual(['v2'], us_bar['signatureVersions'])

    def test_merges_partition_default_keys_with_no_overwrite(self):
        resolver = regions.EndpointResolver(self._template())
        resolved = resolver.construct_endpoint('ec2', 'foo-1')
        self.assertEqual('baz', resolved['foo'])
        self.assertEqual(['http'], resolved['protocols'])

    def test_merges_partition_default_keys_with_overwrite(self):
        resolver = regions.EndpointResolver(self._template())
        resolved = resolver.construct_endpoint('ec2', 'foo-2')
        self.assertEqual('bar', resolved['foo'])
        self.assertEqual(['http'], resolved['protocols'])

    def test_gives_hostname_and_common_name_unaltered(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'eu-baz')
        self.assertEqual('s3.eu-baz.amazonaws.com', result['sslCommonName'])
        self.assertEqual('foo', result['hostname'])

    def tests_uses_partition_endpoint_when_no_region_provided(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized')
        self.assertEqual('not-regionalized', result['hostname'])
        self.assertEqual('aws', result['endpointName'])

    def test_returns_dns_suffix_if_available(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('not-regionalized')
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_dualstack_from_endpoint_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('ec2', 'us-foo',
                                             use_dualstack_endpoint=True)
        self.assertEqual(result['hostname'], 'ec2.foo.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_endpoint_from_service_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('ec2', 'us-bar',
                                             use_dualstack_endpoint=True)
        self.assertEqual(result['hostname'], 'api.ec2.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_endpoint_from_partition_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('dynamodb', 'us-bar',
                                             use_dualstack_endpoint=True)
        self.assertEqual(result['hostname'], 'dynamodb.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_constructs_dualstack_endpoint_no_hostname_in_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'us-fizz',
                                             use_dualstack_endpoint=True)
        self.assertEqual('s3.dualstack.us-fizz.api.aws',
                         result['hostname'])

    def test_constructs_endpoint_dualstack_no_variant_dns_suffix(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'us-bar',
                                             use_dualstack_endpoint=True)
        self.assertEqual('s3.dualstack.us-bar.api.aws',
                         result['hostname'])

    def test_construct_dualstack_endpoint_iso_partition_raise_exception(self):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint('foo', 'foo-1', 'aws-iso',
                                        use_dualstack_endpoint=True)

    def test_get_partition_dns_suffix_no_tags(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual(resolver.get_partition_dns_suffix('aws'),
                         'amazonaws.com')

    def test_get_partition_dualstack_dns_suffix(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertEqual(resolver.get_partition_dns_suffix(
            'aws', ['dualstack']), 'api.aws')

    def test_get_partition_dualstack_dns_suffix_does_not_exist(self):
        resolver = regions.EndpointResolver(self._template())
        self.assertIsNone(resolver.get_partition_dns_suffix(
            'foo', ['dualstack']))

    def test_get_available_fips_endpoints(self):
        resolver = regions.EndpointResolver(self._template())
        fips_endpoints = resolver.get_available_endpoints(
            'ec2', endpoint_variant_tags=['fips'])
        self.assertEqual(fips_endpoints, ['us-foo'])

    def test_get_available_dualstack_endpoints(self):
        resolver = regions.EndpointResolver(self._template())
        dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', endpoint_variant_tags=['dualstack'])
        self.assertEqual(dualstack_endpoints, ['us-foo'])

    def test_get_available_fips_and_dualstack_endpoints(self):
        resolver = regions.EndpointResolver(self._template())
        fips_and_dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', endpoint_variant_tags=['fips', 'dualstack'])
        self.assertEqual(fips_and_dualstack_endpoints, ['us-foo'])

    def test_get_available_fips_endpoints_none(self):
        resolver = regions.EndpointResolver(self._template())
        fips_endpoints = resolver.get_available_endpoints(
            'ec2', 'foo', endpoint_variant_tags=['fips'])
        self.assertEqual(fips_endpoints, [])

    def test_get_available_dualstack_endpoints_none(self):
        resolver = regions.EndpointResolver(self._template())
        dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', 'foo', endpoint_variant_tags=['dualstack'])
        self.assertEqual(dualstack_endpoints, [])

    def test_get_available_fips_and_dualstack_endpoints_none(self):
        resolver = regions.EndpointResolver(self._template())
        fips_and_dualstack_endpoints = resolver.get_available_endpoints(
            'ec2', 'foo', endpoint_variant_tags=['fips', 'dualstack'])
        self.assertEqual(fips_and_dualstack_endpoints, [])

    def test_construct_deprecated_endpoint_raises_warning(self):
        resolver = regions.EndpointResolver(self._template())
        with self.assertLogs('botocore.regions', level='WARNING') as log:
            result = resolver.construct_endpoint('ec2', 'us-dep',
                                                 use_fips_endpoint=True,
                                                 )
            self.assertIn('deprecated endpoint', log.output[0])
            self.assertEqual(result['hostname'], 'ec2-fips.us-dep.amazonaws.com')
            self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_fips_from_endpoint_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('ec2', 'us-foo',
                                             use_fips_endpoint=True)
        self.assertEqual(result['hostname'], 'ec2-fips.foo.amazonaws.com')
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_fips_endpoint_from_service_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('ec2', 'us-bar',
                                             use_fips_endpoint=True)
        self.assertEqual(result['hostname'], 'ec2-fips.us-bar.amazonaws.com')
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_construct_fips_endpoint_from_partition_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('dynamodb', 'us-bar',
                                             use_fips_endpoint=True)
        self.assertEqual(result['hostname'],
                         'dynamodb-fips.us-bar.amazonaws.com')
        self.assertEqual(result['dnsSuffix'], 'amazonaws.com')

    def test_constructs_fips_endpoint_no_hostname_in_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'us-fizz',
                                             use_fips_endpoint=True)
        self.assertEqual('s3-fips.us-fizz.amazonaws.com', result['hostname'])

    def test_construct_dualstack_and_fips_from_endpoint_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('ec2', 'us-foo',
                                             use_dualstack_endpoint=True,
                                             use_fips_endpoint=True)
        self.assertEqual(result['hostname'], 'ec2-fips.foo.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_and_fips_endpoint_from_service_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('ec2', 'us-bar',
                                             use_dualstack_endpoint=True,
                                             use_fips_endpoint=True)
        self.assertEqual(result['hostname'], 'ec2-fips.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_construct_dualstack_and_fips_endpoint_from_partition_default_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('dynamodb', 'us-bar',
                                             use_dualstack_endpoint=True,
                                             use_fips_endpoint=True)
        self.assertEqual(result['hostname'], 'dynamodb-fips.us-bar.api.aws')
        self.assertEqual(result['dnsSuffix'], 'api.aws')

    def test_constructs_dualstack_and_fips_endpoint_no_hostname_in_variant(self):
        resolver = regions.EndpointResolver(self._template())
        result = resolver.construct_endpoint('s3', 'us-fizz',
                                             use_dualstack_endpoint=True,
                                             use_fips_endpoint=True)
        self.assertEqual('s3-fips.dualstack.us-fizz.api.aws',
                         result['hostname'])

    def test_construct_fips_endpoint_no_variant_raise_exception(self):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint('ec2', 'foo-1', 'foo',
                                        use_fips_endpoint=True)

    def test_construct_dualstack_endpoint_no_variant_raise_exception(self):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint('ec2', 'foo-1', 'foo',
                                        use_dualstack_endpoint=True)

    def test_construct_dualstack_and_fips_endpoint_no_variant_raise_exception(self):
        with self.assertRaises(EndpointVariantError):
            resolver = regions.EndpointResolver(self._template())
            resolver.construct_endpoint('ec2', 'foo-1', 'foo',
                                        use_dualstack_endpoint=True,
                                        use_fips_endpoint=True)


def _variant_test_definitions():
    return [
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.us-west-2.amazonaws.com"
        },
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "default-pattern-service-fips.us-west-2.amazonaws.com"
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.af-south-1.amazonaws.com"
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "default-pattern-service-fips.af-south-1.amazonaws.com"
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com"
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": True,
            "dualstack": False,
            "endpoint": "global-service-fips.amazonaws.com"
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com"
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": True,
            "dualstack": False,
            "endpoint": "global-service-fips.amazonaws.com"
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-service.us-west-2.new.dns.suffix"
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-service.af-south-1.new.dns.suffix"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service-fips.us-west-2.new.dns.suffix"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service-fips.af-south-1.new.dns.suffix"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-hostname-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-variant-hostname-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.override-endpoint-variant-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service-fips.af-south-1.amazonaws.com"
        },
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.us-west-2.amazonaws.com"
        },
        {
            "service": "default-pattern-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "default-pattern-service.us-west-2.api.aws"
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "default-pattern-service.af-south-1.amazonaws.com"
        },
        {
            "service": "default-pattern-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "default-pattern-service.af-south-1.api.aws"
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com"
        },
        {
            "service": "global-service",
            "region": "aws-global",
            "fips": False,
            "dualstack": True,
            "endpoint": "global-service.api.aws"
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": False,
            "dualstack": False,
            "endpoint": "global-service.amazonaws.com"
        },
        {
            "service": "global-service",
            "region": "foo",
            "fips": False,
            "dualstack": True,
            "endpoint": "global-service.api.aws"
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-service.dualstack.us-west-2.new.dns.suffix"
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-service.dualstack.af-south-1.new.dns.suffix"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-dns-suffix-service.us-west-2.new.dns.suffix"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-dns-suffix-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-variant-dns-suffix-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-dns-suffix-service.af-south-1.new.dns.suffix"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-hostname-service.dualstack.us-west-2.api.aws"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-variant-hostname-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-variant-hostname-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-variant-hostname-service.dualstack.af-south-1.api.aws"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.us-west-2.amazonaws.com"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-endpoint-variant-service.dualstack.us-west-2.amazonaws.com"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "override-endpoint-variant-service.af-south-1.amazonaws.com"
        },
        {
            "service": "override-endpoint-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "override-endpoint-variant-service.af-south-1.api.aws"
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": False,
            "endpoint": "multi-variant-service.us-west-2.amazonaws.com"
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": False,
            "dualstack": True,
            "endpoint": "multi-variant-service.dualstack.us-west-2.api.aws"
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.multi-variant-service.us-west-2.amazonaws.com"
        },
        {
            "service": "multi-variant-service",
            "region": "us-west-2",
            "fips": True,
            "dualstack": True,
            "endpoint": "fips.multi-variant-service.dualstack.us-west-2.new.dns.suffix"
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": False,
            "endpoint": "multi-variant-service.af-south-1.amazonaws.com"
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": False,
            "dualstack": True,
            "endpoint": "multi-variant-service.dualstack.af-south-1.api.aws"
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": False,
            "endpoint": "fips.multi-variant-service.af-south-1.amazonaws.com"
        },
        {
            "service": "multi-variant-service",
            "region": "af-south-1",
            "fips": True,
            "dualstack": True,
            "endpoint": "fips.multi-variant-service.dualstack.af-south-1.new.dns.suffix"
        }
    ]


def _modeled_variants_template():
    return {
        "partitions" : [
            {
                "defaults" : {
                    "hostname" : "{service}.{region}.{dnsSuffix}",
                    "protocols" : ["https"],
                    "signatureVersions" : ["v4"],
                    "variants" : [
                        {
                            "dnsSuffix" : "amazonaws.com",
                            "hostname" : "{service}-fips.{region}.{dnsSuffix}",
                            "tags" : ["fips"]
                        },
                        {
                            "dnsSuffix" : "api.aws",
                            "hostname" : "{service}.{region}.{dnsSuffix}",
                            "tags" : ["dualstack"]
                        },
                        {
                            "dnsSuffix" : "api.aws",
                            "hostname" : "{service}-fips.{region}.{dnsSuffix}",
                            "tags" : ["dualstack", "fips"]
                        }
                    ]
                },
                "dnsSuffix" : "amazonaws.com",
                "partition" : "aws",
                "regionRegex" : "^(us|eu|ap|sa|ca|me|af)\\-\\w+\\-\\d+$",
                "regions" : {
                    "af-south-1" : {
                        "description" : "Africa (Cape Town)"
                    },
                    "us-west-2" : {
                        "description" : "US West (Oregon)"
                    }
                },
                "services" : {
                    "default-pattern-service" : {
                        "endpoints" : {
                            "af-south-1" : {},
                            "us-west-2" : {
                                "variants" : [
                                    {
                                        "hostname" : "default-pattern-service-fips.us-west-2.amazonaws.com",
                                        "tags" : ["fips"]
                                    },
                                    {
                                        "hostname" : "default-pattern-service.us-west-2.api.aws",
                                        "tags" : ["dualstack"]
                                    }
                                ]
                            }
                        }
                    },
                    "global-service" : {
                        "endpoints" : {
                            "aws-global" : {
                                "credentialScope" : {
                                    "region" : "us-east-1"
                                },
                                "hostname" : "global-service.amazonaws.com",
                                "variants" : [
                                    {
                                        "hostname" : "global-service-fips.amazonaws.com",
                                        "tags" : ["fips"]
                                    },
                                    {
                                        "hostname" : "global-service.api.aws",
                                        "tags" : ["dualstack"]
                                    }
                                ]
                            }
                        },
                        "isRegionalized" : False,
                        "partitionEndpoint" : "aws-global"
                    },
                    "override-variant-service" : {
                        "defaults" : {
                            "variants" : [
                                {
                                    "hostname" : "fips.{service}.{region}.{dnsSuffix}",
                                    "dnsSuffix" : "new.dns.suffix",
                                    "tags" : ["fips"]
                                },
                                {
                                    "hostname" : "{service}.dualstack.{region}.{dnsSuffix}",
                                    "dnsSuffix" : "new.dns.suffix",
                                    "tags" : ["dualstack"]
                                }
                            ]
                        },
                        "endpoints" : {
                            "af-south-1" : {},
                            "us-west-2" : {
                                "variants" : [
                                    {
                                        "hostname" : "fips.override-variant-service.us-west-2.new.dns.suffix",
                                        "tags" : ["fips"]
                                    },
                                    {
                                        "hostname" : "override-variant-service.dualstack.us-west-2.new.dns.suffix",
                                        "tags" : ["dualstack"]
                                    }
                                ]
                            }
                        }
                    },
                    "override-variant-dns-suffix-service" : {
                        "defaults" : {
                            "variants" : [
                                {
                                    "dnsSuffix" : "new.dns.suffix",
                                    "tags" : ["fips"]
                                },
                                {
                                    "dnsSuffix" : "new.dns.suffix",
                                    "tags" : ["dualstack"]
                                }
                            ]
                        },
                        "endpoints" : {
                            "af-south-1" : {},
                            "us-west-2" : {
                                "variants" : [
                                    {
                                        "hostname" : "override-variant-dns-suffix-service-fips.us-west-2.new.dns.suffix",
                                        "tags" : ["fips"]
                                    },
                                    {
                                        "hostname" : "override-variant-dns-suffix-service.us-west-2.new.dns.suffix",
                                        "tags" : ["dualstack"]
                                    }
                                ]
                            }
                        }
                    },
                    "override-variant-hostname-service" : {
                        "defaults" : {
                            "variants" : [
                                {
                                    "hostname" : "fips.{service}.{region}.{dnsSuffix}",
                                    "tags" : ["fips"]
                                },
                                {
                                    "hostname" : "{service}.dualstack.{region}.{dnsSuffix}",
                                    "tags" : ["dualstack"]
                                }
                            ]
                        },
                        "endpoints" : {
                            "af-south-1" : {},
                            "us-west-2" : {
                                "variants" : [
                                    {
                                        "hostname" : "fips.override-variant-hostname-service.us-west-2.amazonaws.com",
                                        "tags" : ["fips"]
                                    },
                                    {
                                        "hostname" : "override-variant-hostname-service.dualstack.us-west-2.api.aws",
                                        "tags" : ["dualstack"]
                                    }
                                ]
                            }
                        }
                    },
                    "override-endpoint-variant-service" : {
                        "endpoints" : {
                            "af-south-1" : {},
                            "us-west-2" : {
                                "variants" : [
                                    {
                                        "hostname" : "fips.override-endpoint-variant-service.us-west-2.amazonaws.com",
                                        "tags" : ["fips"]
                                    },
                                    {
                                        "hostname" : "override-endpoint-variant-service.dualstack.us-west-2.amazonaws.com",
                                        "tags" : ["dualstack"]
                                    }
                                ]
                            }
                        }
                    },
                    "multi-variant-service" : {
                        "defaults" : {
                            "variants" : [
                                {
                                    "hostname" : "fips.{service}.{region}.{dnsSuffix}",
                                    "tags" : ["fips"]
                                },
                                {
                                    "hostname" : "{service}.dualstack.{region}.{dnsSuffix}",
                                    "tags" : ["dualstack"]
                                },
                                {
                                    "dnsSuffix" : "new.dns.suffix",
                                    "hostname" : "fips.{service}.dualstack.{region}.{dnsSuffix}",
                                    "tags" : ["fips", "dualstack"]
                                }
                            ]
                        },
                        "endpoints" : {
                            "af-south-1" : {
                                "deprecated" : True
                            },
                            "us-west-2" : {
                                "variants" : [
                                    {
                                        "hostname" : "fips.multi-variant-service.dualstack.us-west-2.new.dns.suffix",
                                        "tags" : ["fips", "dualstack"]
                                    }
                                ]
                            }
                        }
                    }
                }
            },
            {
                "defaults": {
                    "hostname": "{service}.{region}.{dnsSuffix}",
                    "protocols": ["https"],
                    "signatureVersions": ["v4"]
                },
                "dnsSuffix": "c2s.ic.gov",
                "partition": "aws-iso",
                "regionRegex": "^us\\-iso\\-\\w+\\-\\d+$",
                "regions": {
                    "us-iso-east-1": {
                        "description": "US ISO East"
                    }
                },
                "services": {
                    "some-service": {
                        "endpoints" : {
                            "us-iso-east-1": {}
                        }
                    }
                }
            }
        ],
        "version": 3
    }


@pytest.mark.parametrize("test_case", _variant_test_definitions())
def test_modeled_variants(test_case):
    _verify_expected_endpoint(**test_case)


def _verify_expected_endpoint(service, region, fips, dualstack,
                              endpoint):
    resolver = regions.EndpointResolver(_modeled_variants_template())
    resolved = resolver.construct_endpoint(
        service, region, use_dualstack_endpoint=dualstack,
        use_fips_endpoint=fips
    )
    # If we can't resolve the region, we attempt to get a
    # global endpoint.
    if not resolved:
        resolved = resolver.construct_endpoint(
            service, region, partition_name='aws',
            use_dualstack_endpoint=dualstack,
            use_fips_endpoint=fips
        )
    assert resolved['hostname'] == endpoint


def test_additional_endpoint_data_exists_with_variants():
    resolver = regions.EndpointResolver(_modeled_variants_template())
    resolved = resolver.construct_endpoint(
        'global-service', 'aws-global', use_fips_endpoint=True,
    )
    assert resolved['credentialScope'] == {'region': 'us-east-1'}



import s3transfer
import os
import subprocess
import uuid
import shutil
import argparse
import tempfile


def summarize(script, result_dir, summary_dir):
    """Run the given summary script on every file in the given directory.

    :param script: A summarization script that takes a list of csv files.
    :param result_dir: A directory containing csv performance result files.
    :param summary_dir: The directory to put the summary file in.
    """
    summarize_args = [script]
    for f in os.listdir(result_dir):
        path = os.path.join(result_dir, f)
        if os.path.isfile(path):
            summarize_args.append(path)

    with open(os.path.join(summary_dir, 'summary.txt'), 'wb') as f:
        subprocess.check_call(summarize_args, stdout=f)
    with open(os.path.join(summary_dir, 'summary.json'), 'wb') as f:
        summarize_args.extend(['--output-format', 'json'])
        subprocess.check_call(summarize_args, stdout=f)


def _get_s3transfer_performance_script(script_name):
    """Retrieves an s3transfer performance script if available."""
    s3transfer_directory = os.path.dirname(s3transfer.__file__)
    s3transfer_directory = os.path.dirname(s3transfer_directory)
    scripts_directory = 'scripts/performance'
    scripts_directory = os.path.join(s3transfer_directory, scripts_directory)
    script = os.path.join(scripts_directory, script_name)

    if os.path.isfile(script):
        return script
    else:
        return None


def get_benchmark_script():
    return _get_s3transfer_performance_script('benchmark')


def get_summarize_script():
    return _get_s3transfer_performance_script('summarize')


def backup(source, recursive):
    """Backup a given source to a temporary location.

    :type source: str
    :param source: A local path or s3 path to backup.

    :type recursive: bool
    :param recursive: if True, the source will be treated as a directory.
    """
    if source[:5] == 's3://':
        parts = source.split('/')
        parts.insert(3, str(uuid.uuid4()))
        backup_path = '/'.join(parts)
    else:
        name = os.path.split(source)[-1]
        temp_dir = tempfile.mkdtemp()
        backup_path = os.path.join(temp_dir, name)

    copy(source, backup_path, recursive)
    return backup_path


def copy(source, destination, recursive):
    """Copy files from one location to another.

    The source and destination must both be s3 paths or both be local paths.

    :type source: str
    :param source: A local path or s3 path to backup.

    :type destination: str
    :param destination: A local path or s3 path to backup the source to.

    :type recursive: bool
    :param recursive: if True, the source will be treated as a directory.
    """
    if 's3://' in [source[:5], destination[:5]]:
        cp_args = ['aws', 's3', 'cp', source, destination, '--quiet']
        if recursive:
            cp_args.append('--recursive')
        subprocess.check_call(cp_args)
        return

    if recursive:
        shutil.copytree(source, destination)
    else:
        shutil.copy(source, destination)


def clean(destination, recursive):
    """Delete a file or directory either locally or on S3."""
    if destination[:5] == 's3://':
        rm_args = ['aws', 's3', 'rm', '--quiet', destination]
        if recursive:
            rm_args.append('--recursive')
        subprocess.check_call(rm_args)
    else:
        if recursive:
            shutil.rmtree(destination)
        else:
            os.remove(destination)


def create_random_subfolder(destination):
    """Create a random subdirectory in a given directory."""
    folder_name = str(uuid.uuid4())
    if destination.startswith('s3://'):
        parts = destination.split('/')
        parts.append(folder_name)
        return '/'.join(parts)
    else:
        parts = list(os.path.split(destination))
        parts.append(folder_name)
        path = os.path.join(*parts)
        os.makedirs(path)
        return path


def get_transfer_command(command, recursive, quiet):
    """Get a full cli transfer command.

    Performs common transformations, e.g. adding --quiet
    """
    cli_command = 'aws s3 ' + command

    if recursive:
        cli_command += ' --recursive'

    if quiet:
        cli_command += ' --quiet'
    else:
        print(cli_command)

    return cli_command


def benchmark_command(command, benchmark_script, summarize_script,
                      output_dir, num_iterations, dry_run, upkeep=None,
                      cleanup=None):
    """Benchmark several runs of a long-running command.

    :type command: str
    :param command: The full aws cli command to benchmark

    :type benchmark_script: str
    :param benchmark_script: A benchmark script that takes a command to run
        and outputs performance data to a file. This should be from s3transfer.

    :type summarize_script: str
    :param summarize_script:  A summarization script that the output of the
        benchmark script. This should be from s3transfer.

    :type output_dir: str
    :param output_dir: The directory to output performance results to.

    :type num_iterations: int
    :param num_iterations: The number of times to run the benchmark on the
        command.

    :type dry_run: bool
    :param dry_run: Whether or not to actually run the benchmarks.

    :type upkeep: function that takes no arguments
    :param upkeep: A function that is run after every iteration of the
        benchmark process. This should be used for upkeep, such as restoring
        files that were deleted as part of the command executing.

    :type cleanup: function that takes no arguments
    :param cleanup: A function that is run at the end of the benchmark
        process or if there are any problems during the benchmark process.
        It should be uses for the final cleanup, such as deleting files that
        were created at some destination.
    """
    performance_dir = os.path.join(output_dir, 'performance')
    if os.path.exists(performance_dir):
        shutil.rmtree(performance_dir)
    os.makedirs(performance_dir)

    try:
        for i in range(num_iterations):
            out_file = 'performance%s.csv' % i
            out_file = os.path.join(performance_dir, out_file)
            benchmark_args = [
                benchmark_script, command, '--output-file', out_file
            ]
            if not dry_run:
                subprocess.check_call(benchmark_args)
                if upkeep is not None:
                    upkeep()

        if not dry_run:
            summarize(summarize_script, performance_dir, output_dir)
    finally:
        if not dry_run and cleanup is not None:
            cleanup()


def get_default_argparser():
    """Get an ArgumentParser with all the base benchmark arguments added in."""
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--no-cleanup', action='store_true', default=False,
        help='Do not remove the destination after the tests complete.'
    )
    parser.add_argument(
        '--recursive', action='store_true', default=False,
        help='Indicates that this is a recursive transfer.'
    )
    benchmark_script = get_benchmark_script()
    parser.add_argument(
        '--benchmark-script', default=benchmark_script,
        required=benchmark_script is None,
        help=('The benchmark script to run the commands with. This should be '
              'from s3transfer.')
    )
    summarize_script = get_summarize_script()
    parser.add_argument(
        '--summarize-script', default=summarize_script,
        required=summarize_script is None,
        help=('The summarize script to run the commands with. This should be '
              'from s3transfer.')
    )
    parser.add_argument(
        '-o', '--result-dir', default='results',
        help='The directory to output performance results to. Existing '
             'results will be deleted.'
    )
    parser.add_argument(
        '--dry-run', default=False, action='store_true',
        help='If set, commands will only be printed out, not executed.'
    )
    parser.add_argument(
        '--quiet', default=False, action='store_true',
        help='If set, output is suppressed.'
    )
    parser.add_argument(
        '-n', '--num-iterations', default=1, type=int,
        help='The number of times to run the test.'
    )
    return parser



# Copyright 2013 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import os
import mock

from awscli.customizations.configure import configure, ConfigValue, NOT_SET
from awscli.customizations.configure import profile_to_section
from awscli.testutils import unittest
from awscli.compat import six

from . import FakeSession


class TestConfigureCommand(unittest.TestCase):

    def setUp(self):
        self.writer = mock.Mock()
        self.global_args = mock.Mock()
        self.global_args.profile = None
        self.precanned = PrecannedPrompter(value='new_value')
        self.session = FakeSession({'config_file': 'myconfigfile'})
        self.configure = configure.ConfigureCommand(self.session,
                                                    prompter=self.precanned,
                                                    config_writer=self.writer)

    def assert_credentials_file_updated_with(self, new_values):
        called_args = self.writer.update_config.call_args_list
        credentials_file_call = called_args[0]
        expected_creds_file = os.path.expanduser('~/fake_credentials_filename')
        self.assertEqual(credentials_file_call,
                         mock.call(new_values, expected_creds_file))

    def test_configure_command_sends_values_to_writer(self):
        self.configure(args=[], parsed_globals=self.global_args)
        # Credentials are always written to the shared credentials file.
        self.assert_credentials_file_updated_with(
            {'aws_access_key_id': 'new_value',
             'aws_secret_access_key': 'new_value'})

        # Non-credentials config is written to the config file.
        self.writer.update_config.assert_called_with(
            {'region': 'new_value',
             'output': 'new_value'}, 'myconfigfile')

    def test_same_values_are_not_changed(self):
        # If the user enters the same value as the current value, we don't need
        # to write anything to the config.
        self.configure = configure.ConfigureCommand(self.session,
                                                    prompter=EchoPrompter(),
                                                    config_writer=self.writer)
        self.configure(args=[], parsed_globals=self.global_args)
        self.assertFalse(self.writer.update_config.called)

    def test_none_values_are_not_changed(self):
        # If a user hits enter, this will result in a None value which means
        # don't change the existing values.  In this case, we don't need
        # to write anything out to the config.
        user_presses_enter = None
        precanned = PrecannedPrompter(value=user_presses_enter)
        self.configure = configure.ConfigureCommand(self.session,
                                                    prompter=precanned,
                                                    config_writer=self.writer)
        self.configure(args=[], parsed_globals=self.global_args)
        self.assertFalse(self.writer.update_config.called)

    def test_create_configure_cmd_session_only(self):
        self.configure = configure.ConfigureCommand(self.session)
        self.assertIsInstance(self.configure, configure.ConfigureCommand)

    def test_some_values_changed(self):
        # Test the case where the user only wants to change a single_value.
        responses = {
            "AWS Access Key ID": None,
            "AWS Secert Access Key": None,
            "Default region name": None,
            "Default output format": "NEW OUTPUT FORMAT",
        }
        prompter = KeyValuePrompter(responses)
        self.configure = configure.ConfigureCommand(self.session, prompter=prompter,
                                                    config_writer=self.writer)
        self.configure(args=[], parsed_globals=self.global_args)

        # We only need to write out the default output format.
        self.writer.update_config.assert_called_with(
            {'output': 'NEW OUTPUT FORMAT'}, 'myconfigfile')

    def test_section_name_can_be_changed_for_profiles(self):
        # If the user specifies a profile we need to write this out to
        # the [profile myname] section.
        self.session.profile = 'myname'
        self.configure(args=[], parsed_globals=self.global_args)
        # Note the __section__ key name.
        self.assert_credentials_file_updated_with(
            {'aws_access_key_id': 'new_value',
             'aws_secret_access_key': 'new_value',
             '__section__': 'myname'})
        self.writer.update_config.assert_called_with(
            {'__section__': 'profile myname',
             'region': 'new_value',
             'output': 'new_value'}, 'myconfigfile')

    def test_session_says_profile_does_not_exist(self):
        # Whenever you try to get a config value from botocore,
        # it will raise an exception complaining about ProfileNotFound.
        # We should handle this case, and write out a new profile section
        # in the config file.
        session = FakeSession({'config_file': 'myconfigfile'},
                              profile_does_not_exist=True,
                              profile='profile-does-not-exist')
        self.configure = configure.ConfigureCommand(session,
                                                    prompter=self.precanned,
                                                    config_writer=self.writer)
        self.configure(args=[], parsed_globals=self.global_args)
        self.assert_credentials_file_updated_with(
            {'aws_access_key_id': 'new_value',
             'aws_secret_access_key': 'new_value',
             '__section__': 'profile-does-not-exist'})
        self.writer.update_config.assert_called_with(
            {'__section__': 'profile profile-does-not-exist',
             'region': 'new_value',
             'output': 'new_value'}, 'myconfigfile')


class TestInteractivePrompter(unittest.TestCase):

    def setUp(self):
        self.input_patch = mock.patch('awscli.compat.raw_input')
        self.mock_raw_input = self.input_patch.start()
        self.stdout = six.StringIO()
        self.stdout_patch = mock.patch('sys.stdout', self.stdout)
        self.stdout_patch.start()

    def tearDown(self):
        self.input_patch.stop()
        self.stdout_patch.stop()

    def test_access_key_is_masked(self):
        self.mock_raw_input.return_value = 'foo'
        prompter = configure.InteractivePrompter()
        response = prompter.get_value(
            current_value='myaccesskey', config_name='aws_access_key_id',
            prompt_text='Access key')
        # First we should return the value from raw_input.
        self.assertEqual(response, 'foo')
        # We should also not display the entire access key.
        prompt_text = self.stdout.getvalue()
        self.assertNotIn('myaccesskey', prompt_text)
        self.assertRegex(prompt_text, r'\[\*\*\*\*.*\]')

    def test_access_key_not_masked_when_none(self):
        self.mock_raw_input.return_value = 'foo'
        prompter = configure.InteractivePrompter()
        response = prompter.get_value(
            current_value=None, config_name='aws_access_key_id',
            prompt_text='Access key')
        # First we should return the value from raw_input.
        self.assertEqual(response, 'foo')
        prompt_text = self.stdout.getvalue()
        self.assertIn('[None]', prompt_text)

    def test_secret_key_is_masked(self):
        prompter = configure.InteractivePrompter()
        prompter.get_value(
            current_value='mysupersecretkey',
            config_name='aws_secret_access_key',
            prompt_text='Secret Key')
        # We should also not display the entire secret key.
        prompt_text = self.stdout.getvalue()
        self.assertNotIn('mysupersecretkey', prompt_text)
        self.assertRegex(prompt_text, r'\[\*\*\*\*.*\]')

    def test_non_secret_keys_are_not_masked(self):
        prompter = configure.InteractivePrompter()
        prompter.get_value(
            current_value='mycurrentvalue', config_name='not_a_secret_key',
            prompt_text='Enter value')
        # We should also not display the entire secret key.
        prompt_text = self.stdout.getvalue()
        self.assertIn('mycurrentvalue', prompt_text)
        self.assertRegex(prompt_text, r'\[mycurrentvalue\]')

    def test_user_hits_enter_returns_none(self):
        # If a user hits enter, then raw_input returns the empty string.
        self.mock_raw_input.return_value = ''

        prompter = configure.InteractivePrompter()
        response = prompter.get_value(
            current_value=None, config_name='aws_access_key_id',
            prompt_text='Access key')
        # We convert the empty string to None to indicate that there
        # was no input.
        self.assertIsNone(response)

    def test_compat_input_flushes_after_each_prompt(self):
        # Clear out the default patch
        self.stdout_patch.stop()

        # Create a mock stdout to record flush calls and replace stdout_patch
        self.stdout = mock.Mock()
        self.stdout_patch = mock.patch('sys.stdout', self.stdout)
        self.stdout_patch.start()

        # Make sure flush called at least once
        prompter = configure.InteractivePrompter()
        prompter.get_value(current_value='foo', config_name='bar',
                           prompt_text='baz')
        self.assertTrue(self.stdout.flush.called)

        # Make sure flush is called after *every* prompt
        self.stdout.reset_mock()
        prompter.get_value(current_value='foo2', config_name='bar2',
                           prompt_text='baz2')
        self.assertTrue(self.stdout.flush.called)


class TestConfigValueMasking(unittest.TestCase):

    def test_config_value_is_masked(self):
        config_value = ConfigValue(
            'fake_access_key', 'config_file', 'aws_access_key_id')
        self.assertEqual(config_value.value, 'fake_access_key')
        config_value.mask_value()
        self.assertEqual(config_value.value, '****************_key')

    def test_dont_mask_unset_value(self):
        no_config = ConfigValue(NOT_SET, None, None)
        self.assertEqual(no_config.value, NOT_SET)
        no_config.mask_value()
        self.assertEqual(no_config.value, NOT_SET)


class TestProfileToSection(unittest.TestCase):

    def test_normal_profile(self):
        profile = 'my-profile'
        section = profile_to_section(profile)
        self.assertEqual('profile my-profile', section)

    def test_profile_with_spaces(self):
        profile = 'my spaced profile'
        section = profile_to_section(profile)
        self.assertEqual('profile \'my spaced profile\'', section)

    def test_profile_with_tab(self):
        profile = 'tab\ts'
        section = profile_to_section(profile)
        self.assertEqual('profile \'tab\ts\'', section)

    def test_profile_with_consecutive_spaces(self):
        profile = '    '
        section = profile_to_section(profile)
        self.assertEqual('profile \'    \'', section)

    def test_default_profile(self):
        profile = 'default'
        section = profile_to_section(profile)
        self.assertEqual(profile, section)


class PrecannedPrompter(object):

    def __init__(self, value):
        self._value = value

    def get_value(self, current_value, logical_name, prompt_text=''):
        return self._value


class EchoPrompter(object):

    def get_value(self, current_value, logical_name, prompt_text=''):
        return current_value


class KeyValuePrompter(object):

    def __init__(self, mapping):
        self.mapping = mapping

    def get_value(self, current_value, config_name, prompt_text=''):
        return self.mapping.get(prompt_text)



# Copyright 2012-2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import re

import ruamel.yaml
from ruamel.yaml.resolver import ScalarNode, SequenceNode
from botocore.compat import json
from botocore.compat import OrderedDict


from awscli.compat import six
from awscli.utils import dump_yaml_to_str


def intrinsics_multi_constructor(loader, tag_prefix, node):
    """
    YAML constructor to parse CloudFormation intrinsics.
    This will return a dictionary with key being the intrinsic name
    """

    # Get the actual tag name excluding the first exclamation
    tag = node.tag[1:]

    # Some intrinsic functions doesn't support prefix "Fn::"
    prefix = "Fn::"
    if tag in ["Ref", "Condition"]:
        prefix = ""

    cfntag = prefix + tag

    if tag == "GetAtt" and isinstance(node.value, six.string_types):
        # ShortHand notation for !GetAtt accepts Resource.Attribute format
        # while the standard notation is to use an array
        # [Resource, Attribute]. Convert shorthand to standard format
        value = node.value.split(".", 1)

    elif isinstance(node, ScalarNode):
        # Value of this node is scalar
        value = loader.construct_scalar(node)

    elif isinstance(node, SequenceNode):
        # Value of this node is an array (Ex: [1,2])
        value = loader.construct_sequence(node)

    else:
        # Value of this node is an mapping (ex: {foo: bar})
        value = loader.construct_mapping(node)

    return {cfntag: value}


def _dict_representer(dumper, data):
    return dumper.represent_dict(data.items())


def _add_yaml_1_1_boolean_resolvers(resolver_cls):
    # CloudFormation treats unquoted values that are YAML 1.1 native
    # booleans as booleans, rather than strings. In YAML 1.2, the only
    # boolean values are "true" and "false" so values such as "yes" and "no"
    # when loaded as strings are not quoted when dumped. This logic ensures
    # that we dump these values with quotes so that CloudFormation treats
    # these values as strings and not booleans.
    boolean_regex = re.compile(
        '^(?:yes|Yes|YES|no|No|NO'
        '|true|True|TRUE|false|False|FALSE'
        '|on|On|ON|off|Off|OFF)$'
    )
    boolean_first_chars = list(u'yYnNtTfFoO')
    resolver_cls.add_implicit_resolver(
        'tag:yaml.org,2002:bool', boolean_regex, boolean_first_chars)


def yaml_dump(dict_to_dump):
    """
    Dumps the dictionary as a YAML document
    :param dict_to_dump:
    :return:
    """

    yaml = ruamel.yaml.YAML(typ="safe", pure=True)
    yaml.default_flow_style = False
    yaml.Representer = FlattenAliasRepresenter
    _add_yaml_1_1_boolean_resolvers(yaml.Resolver)
    yaml.Representer.add_representer(OrderedDict, _dict_representer)

    return dump_yaml_to_str(yaml, dict_to_dump)


def _dict_constructor(loader, node):
    # Necessary in order to make yaml merge tags work
    loader.flatten_mapping(node)
    return OrderedDict(loader.construct_pairs(node))


def yaml_parse(yamlstr):
    """Parse a yaml string"""
    try:
        # PyYAML doesn't support json as well as it should, so if the input
        # is actually just json it is better to parse it with the standard
        # json parser.
        return json.loads(yamlstr, object_pairs_hook=OrderedDict)
    except ValueError:
        yaml = ruamel.yaml.YAML(typ="safe", pure=True)
        yaml.Constructor.add_constructor(
            ruamel.yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG,
            _dict_constructor)
        yaml.Constructor.add_multi_constructor(
            "!", intrinsics_multi_constructor)
        _add_yaml_1_1_boolean_resolvers(yaml.Resolver)

        return yaml.load(yamlstr)


class FlattenAliasRepresenter(ruamel.yaml.representer.SafeRepresenter):
    def ignore_aliases(self, data):
        return True



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import BaseAWSCommandParamsTest


class TestPredict(BaseAWSCommandParamsTest):

    prefix = 'machinelearning predict'

    def test_renamed_reboot_arg(self):
        cmdline = self.prefix
        cmdline += ' --ml-model-id ml-foo --record Foo=Bar,Biz=Baz'
        cmdline += ' --predict-endpoint https://myendpoint.amazonaws.com'
        result = {'MLModelId': 'ml-foo',
                  'Record': {'Foo': 'Bar', 'Biz': 'Baz'},
                  'PredictEndpoint': 'https://myendpoint.amazonaws.com'}
        self.assert_params_for_cmd(cmdline, result)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from awscli.testutils import unittest, mock
from awscli.autocomplete.completer import CompletionResult
from awscli.autocomplete import parser
from awscli.autocomplete.serverside.custom_completers.ddb.autocomplete import TableNameCompleter
from tests.unit.autocomplete import InMemoryIndex


class TestTableNameCompleter(unittest.TestCase):
    def setUp(self):
        self.index = InMemoryIndex({
            'command_names': {
                '': [('aws', None)],
                'aws': [('ddb', None)],
                'aws.ddb': [('put', None), ('select', None)],
            },
            'arg_names': {
                '': {
                    'aws': ['region', 'profile'],
                },
                'aws.ddb': {
                    'put': ['table_name'],
                    'select': ['table_name']
                },
            },
            'arg_data': {
                '': {
                    'aws': {
                        'profile': ('profile', 'string', 'aws', '',
                                         None, False, False),
                        'region': ('region', 'string', 'aws', '', None, False,
                                   False),
                    }
                },
                'aws.ddb': {
                    'put': {
                        'table_name': (
                            'table_name', 'string',
                            'put', 'aws.ddb.', None, True, False),
                    },
                    'select': {
                        'table_name': (
                            'table_name', 'string',
                            'select', 'aws.ddb.', None, True, False),
                    },
                }
            }
        })
        self.parser = parser.CLIParser(self.index)
        self.mock_client = mock.Mock()
        self.mock_create_client = mock.Mock()
        self.mock_create_client.create_client.return_value = self.mock_client
        self.completer = TableNameCompleter(self.mock_create_client)

    def test_complete_table_name(self):
        self.mock_client.list_tables.return_value = {
            'TableNames': [
                'tablename',
                'mytable'
            ]
        }
        parsed = self.parser.parse('aws ddb select ')
        results = self.completer.complete(parsed)
        self.assertEqual(
            results,
            [CompletionResult('tablename', 0),
             CompletionResult('mytable', 0)]
        )

    def test_complete_table_name_with_put(self):
        self.mock_client.list_tables.return_value = {
            'TableNames': [
                'tablename',
                'mytable'
            ]
        }
        parsed = self.parser.parse('aws ddb put ')
        results = self.completer.complete(parsed)
        self.assertEqual(
            results,
            [CompletionResult('tablename', 0),
             CompletionResult('mytable', 0)]
        )

    def test_complete_group_name_filters_startswith(self):
        self.mock_client.list_tables.return_value = {
            'TableNames': [
                'tablename',
                'mytable'
            ]
        }
        parsed = self.parser.parse('aws ddb select my')
        results = self.completer.complete(parsed)
        self.assertEqual(
            results,
            [CompletionResult('mytable', -2)]
        )

    def test_complete_group_name_handles_errors(self):
        self.mock_client.list_tables.side_effect = Exception(
            "Something went wrong.")
        parsed = self.parser.parse('aws ddb select ')
        results = self.completer.complete(parsed)
        self.assertEqual(results, [])

    def test_client_created_with_region_and_profiles_from_parsed(self):
        self.mock_client.list_tables.return_value = {'TableNames': []}
        parsed = self.parser.parse(
            'aws --profile foo --region us-west-2 ddb select ')
        self.completer.complete(parsed)
        self.mock_create_client.create_client.assert_called_with(
            'dynamodb', parsed_profile='foo', parsed_region='us-west-2')



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import hashlib
import json
import os
import re

from tests.functional.sso import BaseSSOTest


class TestLoginCommand(BaseSSOTest):
    _TIMESTAMP_FORMAT_PATTERN = re.compile(
        r'\d\d\d\d-\d\d-\d\dT\d\d:\d\d:\d\dZ'
    )

    def add_oidc_workflow_responses(self, access_token,
                                    include_register_response=True):
        responses = [
            # StartDeviceAuthorization response
            {
                'interval': 1,
                'expiresIn': 600,
                'userCode': 'foo',
                'deviceCode': 'foo-device-code',
                'verificationUri': 'https://sso.fake/device',
                'verificationUriComplete': 'https://sso.verify',
            },
            # CreateToken responses
            {
                'Error': {
                    'Code': 'AuthorizationPendingException',
                    'Message': 'Authorization is still pending',
                }
            },
            {
                'expiresIn': self.expires_in,
                'tokenType': 'Bearer',
                'accessToken': access_token,
            }
        ]
        if include_register_response:
            responses.insert(
                0,
                {
                    'clientSecretExpiresAt': self.expiration_time,
                    'clientId': 'foo-client-id',
                    'clientSecret': 'foo-client-secret',
                }
            )
        self.parsed_responses = responses

    def assert_cache_contains_token(
        self,
        start_url,
        expected_token,
        session_name=None,
    ):
        cached_files = os.listdir(self.token_cache_dir)
        # The registration and cached access token
        self.assertEqual(len(cached_files), 2)
        cached_token_filename = self._get_cached_token_filename(
            start_url,
            session_name,
        )
        self.assertIn(cached_token_filename, cached_files)
        self.assertEqual(
            self._get_cached_response(cached_token_filename)['accessToken'],
            expected_token
        )

    def _get_cached_token_filename(self, start_url, session_name):
        to_hash = start_url
        if session_name:
            to_hash = session_name
        return hashlib.sha1(to_hash.encode('utf-8')).hexdigest() + '.json'

    def _get_cached_response(self, token_filename):
        token_path = os.path.join(self.token_cache_dir, token_filename)
        with open(token_path, 'r') as f:
            cached_response = json.loads(f.read())
            return cached_response

    def assert_cache_token_expiration_time_format_is_correct(self):
        token_filename = self._get_cached_token_filename(self.start_url, None)
        token_path = os.path.join(self.token_cache_dir, token_filename)
        with open(token_path, 'r') as f:
            cached_response = json.loads(f.read())
            self.assertIsNotNone(
                self._TIMESTAMP_FORMAT_PATTERN.match(
                    cached_response['expiresAt']
                )
            )

    def test_login(self):
        self.add_oidc_workflow_responses(self.access_token)
        self.run_cmd('sso login')
        self.assert_used_expected_sso_region(expected_region=self.sso_region)
        self.assert_cache_contains_token(
            start_url=self.start_url,
            expected_token=self.access_token
        )

    def test_login_no_browser(self):
        self.add_oidc_workflow_responses(self.access_token)
        stdout, _, _ = self.run_cmd('sso login --no-browser')
        self.assertIn('Browser will not be automatically opened.', stdout)
        self.open_browser_mock.assert_not_called()
        self.assert_used_expected_sso_region(expected_region=self.sso_region)
        self.assert_cache_contains_token(
            start_url=self.start_url,
            expected_token=self.access_token
        )

    def test_login_forces_refresh(self):
        self.add_oidc_workflow_responses(self.access_token)
        self.run_cmd('sso login')
        # The register response from the first login should have been
        # cached.
        self.add_oidc_workflow_responses(
            'new.token', include_register_response=False)
        self.run_cmd('sso login')
        self.assert_cache_contains_token(
            start_url=self.start_url,
            expected_token='new.token'
        )

    def test_login_no_sso_configuration(self):
        self.set_config_file_content(content='')
        _, stderr, _ = self.run_cmd('sso login', expected_rc=253)
        self.assertIn(
            'Missing the following required SSO configuration',
            stderr
        )

    def test_login_minimal_sso_configuration(self):
        content = (
            '[default]\n'
            'sso_start_url={start_url}\n'
            'sso_region={sso_region}\n'
        ).format(start_url=self.start_url, sso_region=self.sso_region)
        self.set_config_file_content(content=content)
        self.add_oidc_workflow_responses(self.access_token)
        self.run_cmd('sso login')
        self.assert_used_expected_sso_region(expected_region=self.sso_region)
        self.assert_cache_contains_token(
            start_url=self.start_url,
            expected_token=self.access_token
        )

    def test_login_partially_missing_sso_configuration(self):
        content = (
            '[default]\n'
            'sso_start_url=%s\n' % self.start_url
        )
        self.set_config_file_content(content=content)
        _, stderr, _ = self.run_cmd('sso login', expected_rc=253)
        self.assertIn(
            'Missing the following required SSO configuration',
            stderr
        )
        self.assertIn('sso_region', stderr)
        self.assertNotIn('sso_start_url', stderr)
        self.assertNotIn('sso_account_id', stderr)
        self.assertNotIn('sso_role_name', stderr)

    def test_token_cache_datetime_format(self):
        self.add_oidc_workflow_responses(self.access_token)
        self.run_cmd('sso login')
        self.assert_used_expected_sso_region(expected_region=self.sso_region)
        self.assert_cache_contains_token(
            start_url=self.start_url,
            expected_token=self.access_token
        )
        self.assert_cache_token_expiration_time_format_is_correct()

    def test_login_sso_session(self):
        content = self.get_sso_session_config('test-session')
        self.set_config_file_content(content=content)
        self.add_oidc_workflow_responses(self.access_token)
        self.run_cmd('sso login')
        self.assert_used_expected_sso_region(expected_region=self.sso_region)
        self.assert_cache_contains_token(
            start_url=self.start_url,
            session_name='test-session',
            expected_token=self.access_token,
        )

    def test_login_sso_with_explicit_sso_session_arg(self):
        content = self.get_sso_session_config(
            'test-session', include_profile=False)
        self.set_config_file_content(content=content)
        self.add_oidc_workflow_responses(self.access_token)
        self.run_cmd('sso login --sso-session test-session')
        self.assert_used_expected_sso_region(expected_region=self.sso_region)
        self.assert_cache_contains_token(
            start_url=self.start_url,
            session_name='test-session',
            expected_token=self.access_token,
        )

    def test_login_sso_session_with_scopes(self):
        self.registration_scopes = ['sso:foo', 'sso:bar']
        content = self.get_sso_session_config('test-session')
        self.set_config_file_content(content=content)
        self.add_oidc_workflow_responses(self.access_token)
        self.run_cmd('sso login')
        self.assert_used_expected_sso_region(expected_region=self.sso_region)
        self.assert_cache_contains_token(
            start_url=self.start_url,
            session_name='test-session',
            expected_token=self.access_token,
        )
        operation, params = self.operations_called[0]
        self.assertEqual(operation.name, 'RegisterClient')
        self.assertEqual(params.get('scopes'), self.registration_scopes)

    def test_login_sso_session_missing_config(self):
        content = (
            f'[default]\n'
            f'sso_session=test\n'
            f'[sso-session test]\n'
        )
        self.set_config_file_content(content=content)
        _, stderr, _ = self.run_cmd('sso login', expected_rc=253)
        self.assertIn(
            'SSO configuration values: sso_start_url, sso_region',
            stderr
        )

    def test_login_sso_session_missing(self):
        content = (
            f'[default]\n'
            f'sso_session=test\n'
        )
        self.set_config_file_content(content=content)
        _, stderr, _ = self.run_cmd('sso login', expected_rc=253)
        self.assertIn('sso-session does not exist: "test"', stderr)



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import argparse
import datetime
import json
import mock

from botocore.exceptions import ClientError

from awscli.customizations import opsworks
from awscli.testutils import unittest
from awscli.customizations.exceptions import ParamValidationError


class TestOpsWorksBase(unittest.TestCase):
    def setUp(self):
        self.mock_session = mock.Mock()
        self.register = opsworks.OpsWorksRegister(self.mock_session)

        # stop the clock while testing
        self.datetime_patcher = mock.patch.object(
            opsworks.datetime, "datetime",
            mock.Mock(wraps=datetime.datetime)
        )
        mocked_datetime = self.datetime_patcher.start()
        mocked_datetime.utcnow.return_value = datetime.datetime(
            2013, 8, 9, 23, 42)

    def tearDown(self):
        self.datetime_patcher.stop()


class TestOpsWorksRegister(TestOpsWorksBase):
    """Tests for functionality independent of the infrastructure class."""

    def _build_args(self, **kwargs):
        return argparse.Namespace(**dict({
            "hostname": None,
            "private_ip": None,
            "public_ip": None,
            "local": False,
            "username": None,
            "private_key": None,
            "ssh": None,
            "target": None,
            "use_instance_profile": False,
        }, **kwargs))

    def test_create_clients_simple(self):
        """Should create clients without additional parameters by default."""

        self.register._create_clients(
            self._build_args(), argparse.Namespace())
        self.mock_session.create_client.assert_has_calls([
            mock.call("iam"),
            mock.call("opsworks"),
        ])

    def test_create_clients_with_region(self):
        """Should pass region names to OpsWorks, but not to IAM clients."""

        self.register._create_clients(
            self._build_args(), argparse.Namespace(region="mars-east-1"))
        self.mock_session.create_client.assert_has_calls([
            mock.call("iam"),
            mock.call("opsworks", region_name="mars-east-1"),
        ])

    def test_create_clients_with_opsworks_endpoint_url(self):
        """Should pass endpoints to OpsWorks, but not to IAM clients."""

        self.register._create_clients(
            self._build_args(), argparse.Namespace(endpoint_url="http://xxx/"))
        self.mock_session.create_client.assert_has_calls([
            mock.call("iam"),
            mock.call("opsworks", endpoint_url="http://xxx/"),
        ])

    def test_create_clients_with_verify_ssl(self):
        """Should pass verify-ssl to OpsWorks, but not to IAM clients."""

        self.register._create_clients(
            self._build_args(), argparse.Namespace(verify_ssl=False))
        self.mock_session.create_client.assert_has_calls([
            mock.call("iam"),
            mock.call("opsworks", verify=False),
        ])

        self.register._create_clients(
            self._build_args(), argparse.Namespace(verify_ssl="/path/to/ca"))
        self.mock_session.create_client.assert_has_calls([
            mock.call("iam"),
            mock.call("opsworks", verify="/path/to/ca"),
        ])

    @mock.patch.object(opsworks, "platform")
    def test_prevalidate_arguments_invalid_hostnames(self, mock_platform):
        """Should only accept valid hostnames."""

        mock_platform.system.return_value = "Linux"
        self.register.prevalidate_arguments(
            self._build_args(
                infrastructure_class="on-premises",
                hostname=None,
                local=True))
        self.register.prevalidate_arguments(
            self._build_args(
                infrastructure_class="on-premises",
                hostname="good-hostname",
                local=True))
        self.register.prevalidate_arguments(
            self._build_args(
                infrastructure_class="on-premises",
                hostname="AlsoAGoodHostname456", local=True))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(
                    infrastructure_class="on-premises",
                    hostname="-bad-hostname",
                    local=True))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(
                    infrastructure_class="on-premises",
                    hostname="f.q.d.n",
                    local=True))

    @mock.patch.object(opsworks, "platform")
    def test_prevalidate_arguments_local_vs_remote(self, mock_platform):
        """Shouldn't allow local and remote mode at the same time."""

        mock_platform.system.return_value = "Linux"
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(self._build_args(
                infrastructure_class="on-premises",
                hostname=None, target=None, local=False))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(self._build_args(
                infrastructure_class="on-premises",
                hostname=None, target="HOSTNAME", local=True))
        self.register.prevalidate_arguments(self._build_args(
            infrastructure_class="on-premises",
            hostname=None, target="HOSTNAME", local=False))
        self.register.prevalidate_arguments(self._build_args(
            infrastructure_class="on-premises",
            hostname=None, target=None, local=True))

        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(self._build_args(
                infrastructure_class="ec2",
                hostname=None, target=None, local=False))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(self._build_args(
                infrastructure_class="ec2",
                hostname=None, target="HOSTNAME", local=True))
        self.register.prevalidate_arguments(self._build_args(
            infrastructure_class="ec2",
            hostname=None, target="HOSTNAME", local=False))
        self.register.prevalidate_arguments(self._build_args(
            infrastructure_class="ec2",
            hostname=None, target=None, local=True))

    @mock.patch.object(opsworks, "platform")
    def test_prevalidate_arguments_local_linux_only(self, mock_platform):
        """Shouldn't allow local and remote mode at the same time."""

        mock_platform.system.return_value = "Linux"
        self.register.prevalidate_arguments(self._build_args(
            infrastructure_class="on-premises", target=None, local=True))
        with self.assertRaises(ParamValidationError):
            mock_platform.system.return_value = "Windows"
            self.register.prevalidate_arguments(self._build_args(
                infrastructure_class="on-premises", target=None, local=True))

    def test_prevalidate_arguments_ssh_override(self):
        """Should not allow --override-ssh and other SSH options."""

        self.register.prevalidate_arguments(self._build_args(
            ssh="telnet", infrastructure_class="ec2", target="i-12345678"
        ))
        self.register.prevalidate_arguments(
            self._build_args(
                username="root", private_key="id_rsa",
                infrastructure_class="ec2", target="1.2.3.4"))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(
                    ssh="telnet", username="root", infrastructure_class="ec2",
                    target="1.2.3.4"))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(
                    ssh="telnet", private_key="id_rsa",
                    infrastructure_class="ec2", target="1.2.3.4"))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(
                    ssh="telnet", username="root", private_key="id_rsa",
                    infrastructure_class="ec2", target="1.2.3.4"))

    def test_create_iam_entities_simple(self):
        """Basic IAM side-effects.

        Should create a group with a user, and an access key."""

        with mock.patch.object(self.register, "iam", create=True) as mock_iam:
            self.register._stack = dict(
                StackId="STACKID", Name="STACKNAME", Arn="ARN")
            self.register._name_for_iam = "HOSTNAME"

            self.register.create_iam_entities(self._build_args())
            policy = "arn:aws:iam::aws:policy/AWSOpsWorksInstanceRegistration"

            mock_iam.create_group.assert_any_call(
                Path="/AWS/OpsWorks/", GroupName="OpsWorks-STACKID")
            mock_iam.create_user.assert_any_call(
                Path="/AWS/OpsWorks/", UserName="OpsWorks-STACKNAME-HOSTNAME")
            mock_iam.add_user_to_group.assert_any_call(
                UserName="OpsWorks-STACKNAME-HOSTNAME",
                GroupName="OpsWorks-STACKID")
            mock_iam.attach_user_policy.assert_any_call(
                PolicyArn=policy,
                UserName="OpsWorks-STACKNAME-HOSTNAME")
            mock_iam.create_access_key.assert_any_call(
                UserName="OpsWorks-STACKNAME-HOSTNAME")

    def test_create_iam_entities_group_exists(self):
        """Should reuse an existing group."""

        with mock.patch.object(self.register, "iam", create=True) as mock_iam:
            self.register._stack = dict(
                StackId="STACKID", Name="STACKNAME", Arn="ARN")
            self.register._name_for_iam = "HOSTNAME"
            mock_iam.create_group.side_effect = ClientError(
                {'Error': {'Code': 'EntityAlreadyExists', 'Message': ''}},
                'CreateGroup')

            self.register.create_iam_entities(self._build_args())

            mock_iam.create_group.assert_any_call(
                Path="/AWS/OpsWorks/", GroupName="OpsWorks-STACKID")
            mock_iam.create_user.assert_any_call(
                Path="/AWS/OpsWorks/", UserName="OpsWorks-STACKNAME-HOSTNAME")
            mock_iam.add_user_to_group.assert_any_call(
                UserName="OpsWorks-STACKNAME-HOSTNAME",
                GroupName="OpsWorks-STACKID")
            mock_iam.create_access_key.assert_any_call(
                UserName="OpsWorks-STACKNAME-HOSTNAME")

    def test_create_iam_entities_user_exists(self):
        """Should use an alternate username if the preferred one is taken."""

        with mock.patch.object(self.register, "iam", create=True) as mock_iam:
            self.register._stack = dict(
                StackId="STACKID", Name="STACKNAME", Arn="ARN")
            self.register._name_for_iam = "HOSTNAME"
            mock_iam.create_user = mock.Mock(
                side_effect=[
                    ClientError(
                        {'Error': {'Code': 'EntityAlreadyExists', 'Message': ''}},
                        'CreateUser'
                    ),
                    ClientError(
                        {'Error': {'Code': 'EntityAlreadyExists', 'Message': ''}},
                        'CreateUser'
                    ),
                    None
                ])

            self.register.create_iam_entities(self._build_args())

            mock_iam.create_group.assert_any_call(
                Path="/AWS/OpsWorks/", GroupName="OpsWorks-STACKID")
            mock_iam.create_user.assert_any_call(
                Path="/AWS/OpsWorks/", UserName="OpsWorks-STACKNAME-HOSTNAME")
            mock_iam.create_user.assert_any_call(
                Path="/AWS/OpsWorks/",
                UserName="OpsWorks-STACKNAME-HOSTNAME+1")
            mock_iam.create_user.assert_any_call(
                Path="/AWS/OpsWorks/",
                UserName="OpsWorks-STACKNAME-HOSTNAME+2")
            mock_iam.add_user_to_group.assert_any_call(
                UserName="OpsWorks-STACKNAME-HOSTNAME+2",
                GroupName="OpsWorks-STACKID")
            mock_iam.create_access_key.assert_any_call(
                UserName="OpsWorks-STACKNAME-HOSTNAME+2")

    def test_create_iam_entities_attach_user_policy_unauthorized(self):
        """Should call PutUserPolicy when AttachUserPolicy is unauthorized."""

        with mock.patch.object(self.register, "iam", create=True) as mock_iam:
            self.register._stack = dict(
                StackId="STACKID", Name="STACKNAME", Arn="ARN")
            self.register._name_for_iam = "HOSTNAME"
            policy = "arn:aws:iam::aws:policy/AWSOpsWorksInstanceRegistration"

            mock_iam.attach_user_policy.side_effect = ClientError(
                {'Error': {'Code': 'AccessDenied', 'Message': ''}},
                'AttachUserPolicy')

            self.register.create_iam_entities(self._build_args())

            mock_iam.attach_user_policy.assert_any_call(
                PolicyArn=policy,
                UserName="OpsWorks-STACKNAME-HOSTNAME")
            mock_iam.put_user_policy.assert_any_call(
                PolicyName="OpsWorks-Instance",
                PolicyDocument=mock.ANY,
                UserName="OpsWorks-STACKNAME-HOSTNAME")

    def test_create_iam_entities_long_names(self):
        """Should shorten IAM entity names to a valid size."""

        long_hostname = "hostname1.very-long-domain-name.within.company.tld"
        shortened_username = \
            "OpsWorks-long-stack-...ork-as-well-hostname1.v...company.tld"

        with mock.patch.object(self.register, "iam", create=True) as mock_iam:
            self.register._stack = dict(
                StackId="STACKID", Name="long stack names should work as well",
                Arn="ARN")
            self.register._name_for_iam = long_hostname

            self.register.create_iam_entities(self._build_args())

            mock_iam.create_group.assert_any_call(
                Path="/AWS/OpsWorks/", GroupName="OpsWorks-STACKID")
            mock_iam.create_user.assert_any_call(
                Path="/AWS/OpsWorks/",
                UserName=shortened_username)
            mock_iam.add_user_to_group.assert_any_call(
                UserName=shortened_username,
                GroupName="OpsWorks-STACKID")
            mock_iam.create_access_key.assert_any_call(
                UserName=shortened_username)

    def test_create_no_iam_entities(self):
        """Should not create IAM entities when using instance profiles."""

        with mock.patch.object(self.register, "iam", create=True) as mock_iam:
            self.register.create_iam_entities(self._build_args(
                use_instance_profile=True
            ))

            self.assertFalse(mock_iam.create_group.called)
            self.assertFalse(mock_iam.create_user.called)
            self.assertFalse(mock_iam.add_user_to_group.called)
            self.assertFalse(mock_iam.create_access_key.called)

    def test_validate_unique_hostname(self):
        """Should detect duplicate host names in the stack early."""

        self.register._stack = {"StackId": "STACKID"}
        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_opsworks.describe_instances.return_value = {
                "Instances": [dict(Hostname="duplicated-hostname")]
            }
            self.register.validate_arguments(
                mock.Mock(hostname="good-hostname"))
            with self.assertRaises(ValueError):
                self.register.validate_arguments(
                    mock.Mock(hostname="duplicated-hostname"))
            with self.assertRaises(ValueError):
                self.register.validate_arguments(
                    mock.Mock(hostname="DUPliCATED-HOSTNAME"))

    @mock.patch.object(opsworks, "tempfile")
    @mock.patch.object(opsworks, "platform")
    @mock.patch.object(opsworks, "subprocess")
    @mock.patch.object(opsworks, "os")
    def test_ssh_windows(
            self, mock_os, mock_subprocess, mock_platform, mock_tempfile):
        """Should use plink on Windows correctly."""

        mock_platform.system.return_value = "Windows"
        self.register._use_address = "ip"
        mock_file = mock.Mock()
        mock_tempfile.NamedTemporaryFile.return_value = mock_file
        mock_file.name = "tmpfilename"

        self.register.ssh(self._build_args(), "script")
        mock_subprocess.check_call.assert_called_with(
            'plink "ip" -m "tmpfilename"', shell=True)

        self.register.ssh(self._build_args(username="foo"), "script")
        mock_subprocess.check_call.assert_called_with(
            'plink -l "foo" "ip" -m "tmpfilename"', shell=True)

        self.register.ssh(self._build_args(private_key="bar"), "script")
        mock_subprocess.check_call.assert_called_with(
            'plink -i "bar" "ip" -m "tmpfilename"', shell=True)

        self.register.ssh(
            self._build_args(username="foo", private_key="bar"),
            "script")
        mock_subprocess.check_call.assert_called_with(
            'plink -l "foo" -i "bar" "ip" -m "tmpfilename"', shell=True)

        self.register.ssh(self._build_args(ssh="plink -agent ip -m"), "script")
        mock_subprocess.check_call.assert_called_with(
            'plink -agent ip -m "tmpfilename"', shell=True)

    @mock.patch.object(opsworks, "platform")
    @mock.patch.object(opsworks, "subprocess")
    def test_ssh_nix(self, mock_subprocess, mock_platform):
        """Should use ssh on non-windows correctly."""

        mock_platform.system.return_value = "Linux"
        self.register._use_address = "ip"

        self.register.ssh(self._build_args(), "script")
        mock_subprocess.check_call.assert_called_with(
            ["ssh", "-tt", "ip", "/bin/sh -c script"])

        self.register.ssh(self._build_args(username="foo"), "script")
        mock_subprocess.check_call.assert_called_with(
            ["ssh", "-tt", "-l", "foo", "ip", "/bin/sh -c script"])

        self.register.ssh(self._build_args(private_key="bar"), "script")
        mock_subprocess.check_call.assert_called_with(
            ["ssh", "-tt", "-i", "bar", "ip", "/bin/sh -c script"])

        self.register.ssh(self._build_args(username="foo", private_key="bar"),
                          "script")
        mock_subprocess.check_call.assert_called_with(
            ["ssh", "-tt", "-l", "foo", "-i", "bar", "ip", "/bin/sh -c script"]
        )
        self.register.ssh(self._build_args(ssh="ssh -k -l foo 1.2.3.4"),
                          "script")
        mock_subprocess.check_call.assert_called_with(
            ["ssh", "-k", "-l", "foo", "1.2.3.4", "/bin/sh -c script"])

    def test_iam_policy_document(self):
        self.assertEqual(
            json.loads(self.register._iam_policy_document("arn::foo")),
            {
                "Statement": [
                    {
                        "Action": "opsworks:RegisterInstance",
                        "Effect": "Allow",
                        "Resource": "arn::foo",
                    }
                ],
                "Version": "2012-10-17"
            }
        )
        self.assertEqual(
            json.loads(self.register._iam_policy_document(
                "arn::foo", datetime.timedelta(minutes=5)
            )), {
                "Statement": [{
                    "Action": "opsworks:RegisterInstance",
                    "Effect": "Allow",
                    "Resource": "arn::foo",
                    "Condition": {
                        "DateLessThan": {
                            "aws:CurrentTime": "2013-08-09T23:47:00Z"
                        }
                    }
                }],
                "Version": "2012-10-17"
            }
        )

    @mock.patch.object(opsworks, "platform")
    @mock.patch.object(opsworks, "subprocess")
    def test_setup_target_machine_remote_nix(
            self, mock_subprocess, mock_platform):
        """Should setup a remote machine from a non-Windows host correctly."""

        mock_platform.system.return_value = "Linux"
        args = self._build_args(
            infrastructure_class="ec2", hostname="HOSTNAME", local=False
        )
        self.register._stack = {"StackId": "STACKID"}
        self.register._prov_params = {
            "AgentInstallerUrl": "URL",
            "Parameters": {"assets_download_bucket": "xxx"}
        }
        self.register.access_key = {
            "AccessKeyId": "AKIAXXX",
            "SecretAccessKey": "foobarbaz"
        }
        self.register._use_address = "ip"
        self.register._use_hostname = "HOSTNAME"

        self.register.setup_target_machine(args)

        cmd = mock_subprocess.check_call.call_args[0][0]
        self.assertEqual(cmd[0], "ssh")
        self.assertEqual(cmd[1], "-tt")
        self.assertEqual(cmd[2], "ip")
        self.assertRegex(cmd[3], r"/bin/sh -c ")

    @mock.patch.object(opsworks, "platform")
    @mock.patch.object(opsworks, "subprocess")
    def test_setup_target_machine_remote_windows(
            self, mock_subprocess, mock_platform):
        """Should setup a remote machine from a Windows host correctly."""

        mock_platform.system.return_value = "Windows"
        args = self._build_args(
            infrastructure_class="ec2", hostname="HOSTNAME", local=False
        )
        self.register._stack = {"StackId": "STACKID"}
        self.register._prov_params = {
            "AgentInstallerUrl": "URL",
            "Parameters": {"assets_download_bucket": "xxx"}
        }
        self.register.access_key = {
            "AccessKeyId": "AKIAXXX",
            "SecretAccessKey": "foobarbaz"
        }
        self.register._use_address = "ip"
        self.register._use_hostname = "HOSTNAME"

        self.register.setup_target_machine(args)

        cmd = mock_subprocess.check_call.call_args[0][0]
        self.assertRegex(cmd, r'^plink ".*" -m ".*"$')

    @mock.patch.object(opsworks, "subprocess")
    def test_setup_target_machine_local(self, mock_subprocess):
        """Should setup the local machine correctly."""

        args = self._build_args(
            infrastructure_class="ec2", local=True
        )
        self.register._stack = {"StackId": "STACKID"}
        self.register._prov_params = {
            "AgentInstallerUrl": "URL",
            "Parameters": {"assets_download_bucket": "xxx"}
        }
        self.register.access_key = {
            "AccessKeyId": "AKIAXXX",
            "SecretAccessKey": "foobarbaz"
        }
        self.register._use_hostname = "HOSTNAME"

        self.register.setup_target_machine(args)

        cmd = mock_subprocess.check_call.call_args[0][0]
        self.assertEqual(cmd[0], "/bin/sh")
        self.assertEqual(cmd[1], "-c")

    def test_pre_config_document_simple(self):
        """Should produce a simple preconfiguration file."""

        self.register._stack = {"StackId": "Foo"}
        self.register._prov_params = {
            "Parameters": {"foo": "Bar", "bar": "Baz"}}
        self.register.access_key = None
        self.register._use_hostname = None

        pre_config = self.register._pre_config_document(
            mock.Mock(private_ip=None, public_ip=None)
        )

        self.assertEqual(
            pre_config, {
                "bar": "Baz",
                "foo": "Bar",
                "import": False,
                "stack_id": "Foo",
            }
        )

    def test_pre_config_document_full(self):
        """Should produce a complex preconfiguration file."""

        self.register._stack = {"StackId": "Foo"}
        self.register._prov_params = {
            "Parameters": {"foo": "Bar", "bar": "Baz"}}
        self.register.access_key = {
            "AccessKeyId": "Bar", "SecretAccessKey": "Baz"}
        self.register._use_hostname = "HOSTNAME"

        pre_config = self.register._pre_config_document(
            mock.Mock(private_ip="PRIVATEIP", public_ip="PUBLICIP"),
        )

        self.assertEqual(
            pre_config, {
                "access_key_id": "Bar",
                "bar": "Baz",
                "foo": "Bar",
                "hostname": "HOSTNAME",
                "import": False,
                "private_ip": "PRIVATEIP",
                "public_ip": "PUBLICIP",
                "secret_access_key": "Baz",
                "stack_id": "Foo",
            }
        )


class TestOpsWorksRegisterEc2(TestOpsWorksBase):
    """Tests for functionality specific to EC2 instances."""

    def _build_args(self, **kwargs):
        return argparse.Namespace(**dict({
            "hostname": None,
            "infrastructure_class": "ec2",
            "private_ip": None,
            "public_ip": None,
            "local": False,
            "username": None,
            "private_key": None,
            "ssh": None,
            "target": None,
            "use_instance_profile": None,
        }, **kwargs))

    @mock.patch.object(opsworks, "subprocess")
    def test_run_main_remote(self, mock_subprocess):
        """Flow test w/ all the expected side-effects for a remote instance."""

        args = self._build_args(stack_id="STACKID", target="i-12345678",
                                local=False)
        parsed_globals = argparse.Namespace()
        mock_ec2 = mock.Mock()
        mock_iam = mock.Mock()
        mock_opsworks = mock.Mock()
        self.mock_session.create_client.side_effect = lambda name, **_: \
            dict(ec2=mock_ec2, iam=mock_iam, opsworks=mock_opsworks)[name]

        mock_opsworks.describe_stacks.return_value = {
            "Stacks": [{
                "Arn": "ARN",
                "Name": "STACKNAME",
                "StackId": "STACKID",
                "Region": "mars-east-1",
            }]
        }
        mock_opsworks.describe_stack_provisioning_parameters.return_value = {
            "AgentInstallerUrl": "URL",
            "Parameters": {
                "assets_download_bucket": "xxx"
            }
        }
        mock_opsworks.describe_instances.return_value = {
            "Instances": []
        }
        mock_ec2.describe_instances.return_value = {
            "Reservations": [{
                "Instances": [{
                    "PublicIpAddress": "192.0.2.42"
                }]
            }]
        }
        mock_iam.create_access_key.return_value = {
            "AccessKey":
                {"AccessKeyId": "AKIAXXX", "SecretAccessKey": "foobarbaz"}
        }

        self.register._run_main(args, parsed_globals)

        mock_iam.create_user.assert_called_with(
            Path="/AWS/OpsWorks/", UserName="OpsWorks-STACKNAME-i-12345678")
        mock_iam.create_group.assert_called_with(
            Path="/AWS/OpsWorks/", GroupName="OpsWorks-STACKID")
        mock_iam.add_user_to_group.assert_called_with(
            UserName="OpsWorks-STACKNAME-i-12345678",
            GroupName="OpsWorks-STACKID")
        mock_iam.create_access_key.assert_called_with(
            UserName="OpsWorks-STACKNAME-i-12345678")
        self.assertTrue(mock_subprocess.check_call.calls)

    @mock.patch.object(opsworks, "urlopen")
    @mock.patch.object(opsworks, "subprocess")
    @mock.patch.object(opsworks, "socket")
    @mock.patch.object(opsworks, "platform")
    def test_run_main_local(
            self, mock_platform, mock_socket, mock_subprocess, mock_urlopen):
        """Flow test w/ all the expected side-effects for a local instance."""

        args = self._build_args(stack_id="STACKID", target=None,
                                local=True)
        parsed_globals = argparse.Namespace()
        mock_ec2 = mock.Mock()
        mock_iam = mock.Mock()
        mock_opsworks = mock.Mock()
        mock_platform.system.return_value = "Linux"
        self.mock_session.create_client.side_effect = lambda name, **_: \
            dict(ec2=mock_ec2, iam=mock_iam, opsworks=mock_opsworks)[name]

        mock_opsworks.describe_stacks.return_value = {
            "Stacks": [{
                "Arn": "ARN",
                "Name": "STACKNAME",
                "StackId": "STACKID",
                "Region": "mars-east-1",
            }]
        }
        mock_opsworks.describe_stack_provisioning_parameters.return_value = {
            "AgentInstallerUrl": "URL",
            "Parameters": {
                "assets_download_bucket": "xxx"
            }
        }
        mock_opsworks.describe_instances.return_value = {
            "Instances": []
        }
        mock_iam.create_access_key.return_value = {
            "AccessKey":
                {"AccessKeyId": "AKIAXXX", "SecretAccessKey": "foobarbaz"}
        }
        mock_socket.gethostname.return_value = "HOSTNAME"
        mock_urlopen.return_value.read.return_value = \
            '{"region": "mars-east-1"}'

        self.register._run_main(args, parsed_globals)

        mock_iam.create_user.assert_called_with(
            Path="/AWS/OpsWorks/", UserName="OpsWorks-STACKNAME-HOSTNAME")
        mock_iam.create_group.assert_called_with(
            Path="/AWS/OpsWorks/", GroupName="OpsWorks-STACKID")
        mock_iam.add_user_to_group.assert_called_with(
            UserName="OpsWorks-STACKNAME-HOSTNAME",
            GroupName="OpsWorks-STACKID")
        mock_iam.create_access_key.assert_called_with(
            UserName="OpsWorks-STACKNAME-HOSTNAME")
        self.assertTrue(mock_subprocess.check_call.calls)

    def test_prevalidate_arguments_no_ips_for_ec2(self):
        """Shouldn't allow overriding IP addresses for EC2."""

        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(
                    target="target", private_ip="private-ip"))
        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(target="target", public_ip="public-ip"))

    @mock.patch.object(opsworks, "urlopen")
    def test_validate_same_region(self, mock_urlopen):
        """Should ensure that the local instance is in the correct region."""

        self.register._stack = {"Region": "mars-east-1"}

        mock_urlopen.return_value.read.return_value = \
            '{"region": "mars-east-1"}'
        self.register.validate_arguments(
            self._build_args(hostname=None, local=True))

        with self.assertRaises(ValueError):
            mock_urlopen.return_value.read.return_value = \
                '{"region": "mars-west-1"}'
            self.register.validate_arguments(
                self._build_args(hostname=None, local=True))

    @mock.patch.object(opsworks, "urlopen")
    def test_validate_same_region_bytes(self, mock_urlopen):
        """Check that register can handle bytes returned by urlopen.read"""

        self.register._stack = {"Region": "mars-east-1"}

        mock_urlopen.return_value.read.return_value = \
            b'{"region": "mars-east-1"}'
        try:
            self.register.validate_arguments(
                self._build_args(hostname=None, local=True))
        except Exception as e:
            self.fail(
                'Register should work with bytes response from urlopen.read. '
                'Got exception: %s' % e)

    def test_retrieve_stack_ec2(self):
        """Should retrieve an EC2 stack and the matching instance."""

        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_ec2 = mock.Mock()
            self.mock_session.create_client.return_value = mock_ec2
            mock_opsworks.describe_stacks.return_value = {
                "Stacks": [{
                    "StackId": "STACKID",
                    "Region": "mars-east-1",
                }]
            }
            mock_ec2.describe_instances.return_value = {
                "Reservations": [{
                    "Instances": [{
                        "InstanceId": "i-12345678",
                    }]
                }, {
                    "Instances": [{
                        "InstanceId": "i-9abcdef0",
                        "VpcId": "vpc-123456"
                    }]
                }
                ]
            }
            self.register.retrieve_stack(self._build_args(
                stack_id="STACKID", target="i-12345678"
            ))
            mock_ec2.describe_instances.assert_called_with(
                InstanceIds=["i-12345678"], Filters=[]
            )
            self.assertEqual(
                self.register._ec2_instance["InstanceId"],
                "i-12345678")

    def test_retrieve_stack_vpc(self):
        """Should retrieve an VPC stack and the matching instance."""

        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_ec2 = mock.Mock()
            self.mock_session.create_client.return_value = mock_ec2
            mock_opsworks.describe_stacks.return_value = {
                "Stacks": [{
                    "StackId": "STACKID",
                    "Region": "mars-east-1",
                    "VpcId": "vpc-123456",
                }]
            }
            mock_ec2.describe_instances.return_value = {
                "Reservations": [{
                    "Instances": [{
                        "InstanceId": "i-12345678",
                    }]
                }]
            }
            self.register.retrieve_stack(self._build_args(
                stack_id="STACKID", target="i-12345678"
            ))
            mock_ec2.describe_instances.assert_called_with(
                InstanceIds=["i-12345678"],
                Filters=[{
                    "Name": "vpc-id",
                    "Values": ["vpc-123456"]
                }])
            self.assertEqual(
                self.register._ec2_instance["InstanceId"],
                "i-12345678")

    def test_retrieve_stack_ec2_instance_id(self):
        """Should find an EC2 instance by instance ID."""

        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_ec2 = mock.Mock()
            self.mock_session.create_client.return_value = mock_ec2
            mock_opsworks.describe_stacks.return_value = {
                "Stacks": [{
                    "StackId": "STACKID",
                    "Region": "mars-east-1",
                }]
            }
            mock_ec2.describe_instances.return_value = {
                "Reservations": [{
                    "Instances": [{
                        "InstanceId": "i-12345678910",
                    }]
                }]
            }
            self.register.retrieve_stack(self._build_args(
                stack_id="STACKID", target="i-12345678910"
            ))
            mock_ec2.describe_instances.assert_called_with(
                InstanceIds=["i-12345678910"], Filters=[]
            )
            self.assertEqual(
                self.register._ec2_instance["InstanceId"],
                "i-12345678910")

    def test_retrieve_stack_target_ip_address(self):
        """Should find an EC2 instance by IP address."""

        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_ec2 = mock.Mock()
            self.mock_session.create_client.return_value = mock_ec2
            mock_opsworks.describe_stacks.return_value = {
                "Stacks": [{
                    "StackId": "STACKID",
                    "Region": "mars-east-1",
                }]
            }
            mock_ec2.describe_instances.return_value = {
                "Reservations": [{
                    "Instances": [{
                        "InstanceId": "i-12345678",
                        "PrivateIpAddress": "1.2.3.4"
                    }, {
                        "InstanceId": "i-9abcdef0",
                        "PrivateIpAddress": "1.2.3.5"
                    }]
                }]
            }
            self.register.retrieve_stack(self._build_args(
                stack_id="STACKID", target="1.2.3.4"
            ))
            mock_ec2.describe_instances.assert_called_with(Filters=[])
            self.assertEqual(
                self.register._ec2_instance["InstanceId"],
                "i-12345678")

    def test_retrieve_stack_target_name(self):
        """Should find an EC2 instance by name."""

        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_ec2 = mock.Mock()
            self.mock_session.create_client.return_value = mock_ec2
            mock_opsworks.describe_stacks.return_value = {
                "Stacks": [{
                    "StackId": "STACKID",
                    "Region": "mars-east-1",
                }]
            }
            mock_ec2.describe_instances.return_value = {
                "Reservations": [{
                    "Instances": [{
                        "InstanceId": "i-12345678",
                        "PrivateIpAddress": "1.2.3.4"
                    }]
                }]
            }
            self.register.retrieve_stack(self._build_args(
                stack_id="STACKID", target="db-master"
            ))
            mock_ec2.describe_instances.assert_called_with(
                Filters=[{"Name": "tag:Name", "Values": ["db-master"]}]
            )
            self.assertEqual(
                self.register._ec2_instance["InstanceId"],
                "i-12345678")

    def test_retrieve_stack_target_none(self):
        """Should complain if it cannot find matching instances."""

        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_ec2 = mock.Mock()
            self.mock_session.create_client.return_value = mock_ec2
            mock_opsworks.describe_stacks.return_value = {
                "Stacks": [{
                    "StackId": "STACKID",
                    "Region": "mars-east-1",
                }]
            }
            mock_ec2.describe_instances.return_value = {
                "Reservations": []
            }
            with self.assertRaises(ValueError):
                self.register.retrieve_stack(
                    self._build_args(
                        stack_id="STACKID", target="some-instance"))

    def test_retrieve_stack_target_too_many(self):
        """Should complain if it finds too many matching instances."""

        with mock.patch.object(
                self.register, "opsworks", create=True) as mock_opsworks:
            mock_ec2 = mock.Mock()
            self.mock_session.create_client.return_value = mock_ec2
            mock_opsworks.describe_stacks.return_value = {
                "Stacks": [{
                    "StackId": "STACKID",
                    "Region": "mars-east-1",
                }]
            }
            mock_ec2.describe_instances.return_value = {
                "Reservations": [{
                    "Instances": [{
                        "InstanceId": "i-12345678",
                        "PrivateIpAddress": "1.2.3.4"
                    }, {
                        "InstanceId": "i-9abcdef0",
                        "PrivateIpAddress": "1.2.3.5"
                    }]
                }]
            }
            with self.assertRaises(ValueError):
                self.register.retrieve_stack(self._build_args(
                    stack_id="STACKID", target="some-instance"
                ))

    def test_determine_details_simple(self):
        """Should determine names and address for a basic EC2 instance."""

        self.register._ec2_instance = {
            "PublicIpAddress": "192.0.2.42"
        }
        self.register._use_address = None
        self.register.determine_details(self._build_args(
            target="i-12345678"
        ))
        self.assertEqual(self.register._use_address, "192.0.2.42")
        self.assertEqual(self.register._use_hostname, None)
        self.assertEqual(self.register._name_for_iam, "i-12345678")

    def test_determine_details_with_hostname(self):
        """Should determine names and address with a hostname override."""

        self.register._ec2_instance = {
            "PublicIpAddress": "192.0.2.42"
        }
        self.register._use_address = None
        self.register.determine_details(self._build_args(
            infrastructure_class="ec2",
            hostname="prettyhostname"
        ))
        self.assertEqual(self.register._use_address, "192.0.2.42")
        self.assertEqual(self.register._use_hostname, "prettyhostname")
        self.assertEqual(self.register._name_for_iam, "prettyhostname")

    def test_determine_details_private_ip_only(self):
        """Should determine names and address for a EC2 instance without a
        public IP address."""

        self.register._ec2_instance = {
            "PrivateIpAddress": "192.0.2.42"
        }
        self.register._use_address = None
        self.register.determine_details(self._build_args(
            target="i-12345678"
        ))
        self.assertEqual(self.register._use_address, "192.0.2.42")
        self.assertEqual(self.register._use_hostname, None)
        self.assertEqual(self.register._name_for_iam, "i-12345678")

    @mock.patch.object(opsworks, "socket")
    def test_determine_details_local_simple(self, mock_socket):
        """Should determine names and address for the local EC2 instance."""

        mock_socket.gethostname.return_value = "HOSTNAME"
        self.register._use_address = None
        self.register.determine_details(self._build_args(
            hostname=None, local=True
        ))
        self.assertEqual(self.register._use_address, None)
        self.assertEqual(self.register._use_hostname, None)
        self.assertEqual(self.register._name_for_iam, "HOSTNAME")

    def test_determine_details_local_with_hostname(self):
        """Should determine names and address for the local EC2 instance with
        a hostname override."""

        self.register._use_address = None
        self.register.determine_details(self._build_args(
            hostname="prettyhostname", local=True
        ))
        self.assertEqual(self.register._use_address, None)
        self.assertEqual(self.register._use_hostname, "prettyhostname")
        self.assertEqual(self.register._name_for_iam, "prettyhostname")

    def test_determine_details_given_address(self):
        """Should use a given address."""

        self.register._use_address = "192.0.2.42"
        self.register.determine_details(self._build_args(
            target="192.0.2.42"
        ))
        self.assertEqual(self.register._use_address, "192.0.2.42")
        self.assertEqual(self.register._use_hostname, None)
        self.assertEqual(self.register._name_for_iam, "192.0.2.42")


class TestOpsWorksRegisterOnPremises(TestOpsWorksBase):
    """Tests for functionality specific to on-premises instances."""

    def _build_args(self, **kwargs):
        return argparse.Namespace(**dict({
            "hostname": None,
            "infrastructure_class": "on-premises",
            "private_ip": None,
            "public_ip": None,
            "local": False,
            "username": None,
            "private_key": None,
            "ssh": None,
            "target": None,
            "use_instance_profile": None,
        }, **kwargs))

    @mock.patch.object(opsworks, "subprocess")
    def test_run_main(self, mock_subprocess):
        """Flow test w/ all the expected side-effects for a remote instance."""

        args = self._build_args(stack_id="STACKID", target="HOSTNAME",
                                local=False, ssh=None, hostname=None,
                                private_ip=None, public_ip=None)
        parsed_globals = argparse.Namespace()
        mock_ec2 = mock.Mock()
        mock_iam = mock.Mock()
        mock_opsworks = mock.Mock()
        self.mock_session.create_client.side_effect = lambda name, **_: \
            dict(ec2=mock_ec2, iam=mock_iam, opsworks=mock_opsworks)[name]

        mock_opsworks.describe_stacks.return_value = {
            "Stacks": [{
                "Arn": "ARN",
                "Name": "STACKNAME",
                "StackId": "STACKID",
                "Region": "mars-east-1",
            }]
        }
        mock_opsworks.describe_stack_provisioning_parameters.return_value = {
            "AgentInstallerUrl": "URL",
            "Parameters": {
                "assets_download_bucket": "xxx"
            }
        }
        mock_opsworks.describe_instances.return_value = {
            "Instances": []
        }
        mock_iam.create_access_key.return_value = {
            "AccessKey":
            {"AccessKeyId": "AKIAXXX", "SecretAccessKey": "foobarbaz"}}

        self.register._run_main(args, parsed_globals)

        mock_iam.create_user.assert_called_with(
            Path="/AWS/OpsWorks/", UserName="OpsWorks-STACKNAME-HOSTNAME")
        mock_iam.create_group.assert_called_with(
            Path="/AWS/OpsWorks/", GroupName="OpsWorks-STACKID")
        mock_iam.add_user_to_group.assert_called_with(
            UserName="OpsWorks-STACKNAME-HOSTNAME",
            GroupName="OpsWorks-STACKID")
        mock_iam.create_access_key.assert_called_with(
            UserName="OpsWorks-STACKNAME-HOSTNAME")
        self.assertTrue(mock_subprocess.check_call.calls)

    def test_prevalidate_arguments_no_instance_profile(self):
        """Shouldn't allow using an instance profile on-premises."""

        with self.assertRaises(ParamValidationError):
            self.register.prevalidate_arguments(
                self._build_args(
                    target="target", use_instance_profile=True))

    def test_determine_details_simple(self):
        """Should determine names and address for a basic instance."""

        self.register._use_address = None
        self.register.determine_details(self._build_args(
            infrastructure_class="on-premises",
            target="HOSTNAME", hostname=None, local=False
        ))
        self.assertEqual(self.register._use_address, "HOSTNAME")
        self.assertEqual(self.register._use_hostname, None)
        self.assertEqual(self.register._name_for_iam, "HOSTNAME")

    def test_determine_details_with_hostname(self):
        """Should determine names and address with a hostname override."""

        self.register._use_address = None
        self.register.determine_details(self._build_args(
            infrastructure_class="on-premises",
            target="HOSTNAME", hostname="prettyhostname", local=False
        ))
        self.assertEqual(self.register._use_address, "HOSTNAME")
        self.assertEqual(self.register._use_hostname, "prettyhostname")
        self.assertEqual(self.register._name_for_iam, "prettyhostname")

    @mock.patch.object(opsworks, "socket")
    def test_determine_details_local_simple(self, mock_socket):
        """Should determine names and address for the local instance."""

        mock_socket.gethostname.return_value = "HOSTNAME"
        self.register._use_address = None
        self.register.determine_details(self._build_args(
            infrastructure_class="on-premises",
            hostname=None, local=True
        ))
        self.assertEqual(self.register._use_address, None)
        self.assertEqual(self.register._use_hostname, None)
        self.assertEqual(self.register._name_for_iam, "HOSTNAME")

    def test_determine_details_local_with_hostname(self):
        """Should determine names and address for the local instance with a
        hostname override."""

        self.register._use_address = None
        self.register.determine_details(self._build_args(
            infrastructure_class="on-premises",
            hostname="prettyhostname", local=True
        ))
        self.assertEqual(self.register._use_address, None)
        self.assertEqual(self.register._use_hostname, "prettyhostname")
        self.assertEqual(self.register._name_for_iam, "prettyhostname")


class TestOpsWorksHelpers(unittest.TestCase):
    """Tests for helper functions."""

    def test_clean_for_iam(self):
        """Should sanitize strings for IAM."""

        self.assertEqual(opsworks.clean_for_iam("foobar"), "foobar")
        self.assertEqual(opsworks.clean_for_iam("foo bar 123"), "foo-bar-123")
        self.assertEqual(opsworks.clean_for_iam("baz&@%#^*$bar"), "baz-@-bar")

    def test_shorten_name(self):
        """Should shorten strings by introducing ellipses."""

        # short
        self.assertEqual(opsworks.shorten_name("1234", 5), "1234")
        self.assertEqual(opsworks.shorten_name("12345", 5), "12345")
        # odd number of characters
        self.assertEqual(opsworks.shorten_name("123456789", 5), "1...9")
        self.assertEqual(opsworks.shorten_name("123456789", 6), "12...9")
        self.assertEqual(opsworks.shorten_name("123456789", 7), "12...89")
        self.assertEqual(opsworks.shorten_name("123456789", 8), "123...89")
        self.assertEqual(opsworks.shorten_name("123456789", 9), "123456789")
        self.assertEqual(opsworks.shorten_name("123456789", 10), "123456789")
        # even number of characters
        self.assertEqual(opsworks.shorten_name("1234567890", 5), "1...0")
        self.assertEqual(opsworks.shorten_name("1234567890", 6), "12...0")
        self.assertEqual(opsworks.shorten_name("1234567890", 7), "12...90")
        self.assertEqual(opsworks.shorten_name("1234567890", 8), "123...90")
        self.assertEqual(opsworks.shorten_name("1234567890", 9), "123...890")
        self.assertEqual(opsworks.shorten_name("1234567890", 10), "1234567890")
        self.assertEqual(opsworks.shorten_name("1234567890", 11), "1234567890")



# Copyright 2019 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
import logging
import shlex
import sys
from contextlib import nullcontext, contextmanager

from prompt_toolkit.application import Application
from prompt_toolkit.completion import Completer, ThreadedCompleter
from prompt_toolkit.completion import Completion
from prompt_toolkit.document import Document

from awscli.logger import LOG_FORMAT, disable_crt_logging
from awscli.autocomplete import parser
from awscli.autocomplete.local import model
from awscli.autoprompt.doc import DocsGetter
from awscli.autoprompt.output import OutputGetter
from awscli.autoprompt.factory import PromptToolkitFactory
from awscli.autoprompt.logger import PromptToolkitHandler


LOG = logging.getLogger(__name__)


@contextmanager
def loggers_handler_switcher():
    old_handlers = {}
    loggers = [name for name in logging.root.manager.loggerDict]

    handler = PromptToolkitHandler()
    handler.setLevel(logging.DEBUG)
    handler.setFormatter(logging.Formatter(LOG_FORMAT))
    for logger_name in loggers:
        logger = logging.getLogger(logger_name)
        old_handlers[logger_name] = logger.handlers
        logger.handlers = []
        logger.addHandler(handler)
    yield None
    for logger_name in loggers:
        logger = logging.getLogger(logger_name)
        logger.handlers = []
        for handler in old_handlers[logger_name]:
            logger.addHandler(handler)


class PromptToolkitPrompter:
    """Handles the actual prompting in the autoprompt workflow.

    """
    def __init__(self, completion_source, driver, completer=None,
                 factory=None, app=None, cli_parser=None, output=None,
                 app_input=None):
        self._completion_source = completion_source
        self._output = output
        self._input = app_input
        if completer is None:
            completer = PromptToolkitCompleter(self._completion_source)
        # We wrap our completer with a ThreadedCompleter to make autocompletion
        # generation more responsive.
        self._completer = ThreadedCompleter(completer)
        if factory is None:
            factory = PromptToolkitFactory(completer=self._completer)
        self._parser = cli_parser
        if self._parser is None:
            self._parser = parser.CLIParser(
                model.ModelIndex(), return_first_command_match=True)
        self._factory = factory
        self.input_buffer = None
        self.doc_buffer = None
        self.output_buffer = None
        if app is None:
            app = self.create_application()
        self.app = app
        self._args = []
        self._driver = driver
        self._docs_getter = DocsGetter(self._driver)
        self._output_getter = OutputGetter(self._driver)

    def args(self, value):
        self._args = value

    args = property(None, args)

    def _create_buffers(self):
        self.input_buffer = self._factory.create_input_buffer(
            self.update_bottom_buffers_text)
        self.doc_buffer = self._factory.create_doc_buffer()
        self.output_buffer = self._factory.create_output_buffer()

    def _create_containers(self):
        input_buffer_container = self._factory.create_input_buffer_container(
            self.input_buffer)
        doc_window = self._factory.create_searchable_window(
            'Doc panel', self.doc_buffer)
        output_window = self._factory.create_searchable_window(
            'Output panel', self.output_buffer)
        return input_buffer_container, doc_window, output_window

    def create_application(self):
        self._create_buffers()
        input_buffer_container, \
                doc_window, output_window = self._create_containers()
        layout = self._factory.create_layout(
            on_input_buffer_text_changed=self.update_bottom_buffers_text,
            input_buffer_container=input_buffer_container,
            doc_window=doc_window, output_window=output_window
        )
        kb_manager = self._factory.create_key_bindings()
        kb = kb_manager.keybindings
        app = Application(layout=layout, key_bindings=kb, full_screen=False,
                          output=self._output, erase_when_done=True,
                          input=self._input)
        self._set_app_defaults(app)
        return app

    def _set_app_defaults(self, app):
        app.show_doc = False
        app.multi_column = False
        app.show_help = False
        app.debug = False
        app.show_output = False
        return app

    def update_bottom_buffers_text(self, *args):
        parsed = self._parser.parse(
            'aws ' + self.input_buffer.document.text)
        self._update_doc_window_contents(parsed)
        self._update_output_window_contents(parsed)

    def _update_output_window_contents(self, parsed):
        content = self._output_getter.get_output(parsed)
        if content is not None and content != self.output_buffer.document.text:
            self.output_buffer.reset()
            new_document = Document(text=content, cursor_position=0)
            self.output_buffer.set_document(new_document, bypass_readonly=True)

    def _update_doc_window_contents(self, parsed):
        content = self._docs_getter.get_docs(parsed)
        # The only way to "modify" a read-only buffer in prompt_toolkit is to
        # create a new `prompt_toolkit.document.Document`. A 'cursor_position'
        # of 0 allows us to start viewing the documentation from the beginning,
        # as opposed to the default of moving the cursor to the end.
        #
        # Note: `content` may be None if we are still in the middle of typing a
        # command (e.g. `aws ec2 descr`) or if the current command is invalid.
        # In these cases, we don't update the doc window.
        if content is not None and content != self.doc_buffer.document.text:
            self.doc_buffer.reset()
            new_document = Document(text=content, cursor_position=0)
            self.doc_buffer.set_document(new_document, bypass_readonly=True)

    def _set_debug_mode(self):
        if '--debug' in self._args:
            self.app.debug = True

    def pre_run(self):
        if '--cli-auto-prompt' in self._args:
            self._args.remove('--cli-auto-prompt')
        # Before prompting the user for a command, we copy what they entered
        # at the command line into the buffer (aka "construction zone"), set
        # the cursor position, and make sure to immediately provide a pop-up
        # menu with autocompletion options.
        #
        # If what the user entered at the command line already has no
        # completions, we just keep it as it is without showing
        # any suggestions.
        self.input_buffer.insert_text(' '.join(self._args))
        cmd_line_text = 'aws ' + self.input_buffer.document.text
        self._set_input_buffer_text(cmd_line_text)
        self.update_bottom_buffers_text()
        self.input_buffer.start_completion()

    def _set_input_buffer_text(self, cmd_line_text):
        """If entered command line does not have trailing space and can not
           be autocompleted we assume that it is a completed part of command
           and add trailing space to it"""
        if cmd_line_text[-1] == ' ':
            return
        if self._can_autocomplete(cmd_line_text):
            return
        if cmd_line_text.strip() != 'aws':
            self.input_buffer.insert_text(' ')

    def _can_autocomplete(self, cmd_line_text):
        return bool(self._completion_source.autocomplete(cmd_line_text))

    def _quote_args_with_spaces(self, args):
        return [shlex.quote(arg) for arg in args]

    def prompt_for_args(self, original_args):
        """Prompt for values for a given CLI command.

        :type original_args: list
        :param original_args: A list of the entered command at the program entry
            point, less the initial `aws` command.

        :rtype: list
        :return: A list of what was entered into the autoprompt buffer,
            delimited by whitespace characters.

        """
        self._args = self._quote_args_with_spaces(original_args)
        self._set_debug_mode()
        logging_manager = nullcontext
        if self.app.debug:
            # NOTE: The CRT library is not integrated with the Python stdlib
            # logger and only allows you to write to file names and
            # stdout/stderr. Therefore we currently are unable to switch it
            # to write logs to our promptoolkit debug buffer. So instead of
            # letting it write to stderr and disrupting the UI, we are
            # disabling it while in auto-prompt mode until it has
            # integrations with the Python stdlib logger.
            disable_crt_logging()
            logging_manager = loggers_handler_switcher
        with logging_manager():
            self.app.run(pre_run=self.pre_run)
        cmd_line_text = self.input_buffer.document.text
        # Once the application is finished running, the screen is cleared.
        # Here, we display the command to be run so that the user knows what
        # command caused the output that they're seeing.
        sys.stdout.write(f'> aws {cmd_line_text}\n')
        return shlex.split(cmd_line_text)


class PromptToolkitCompleter(Completer):
    """Handles conversion of autocompletion data from our completion source
    into `prompt_toolkit.Completion` objects that prompt_toolkit understands.

    Note: prompt_toolkit expects a method `get_completions()` to generate the
    `prompt_toolkit.Completion` objects.

    """
    def __init__(self, completion_source):
        self._completion_source = completion_source

    def _convert_to_prompt_completions(self, low_level_completions,
                                       text_before_cursor):
        # Converts the low-level completions from the model autocompleter
        # and converts them to Completion() objects that are used by
        # prompt_toolkit.
        word_before_cursor = self._strip_whitespace(text_before_cursor)
        completions = self._filter_completions(low_level_completions)
        ordered_completions = self._prioritize_required_args(completions)
        for completion in ordered_completions:
            display_text = self._get_display_text(completion)
            display_meta = self._get_display_meta(completion)
            location = self._get_starting_location_of_last_word(
                text_before_cursor, word_before_cursor)
            yield Completion(completion.name, location, display=display_text,
                             display_meta=display_meta)

    def get_completions(self, document, complete_event):
        try:
            text_before_cursor = document.text_before_cursor
            text_to_autocomplete = 'aws ' + text_before_cursor
            completions = self._completion_source.autocomplete(
                text_to_autocomplete, len(text_to_autocomplete))
            yield from self._convert_to_prompt_completions(
                completions, text_before_cursor)
        except Exception as e:
            LOG.debug('Exception caught in PromptToolkitCompleter: %s' % e,
                      exc_info=True)

    def _strip_whitespace(self, text):
        word_before_cursor = ''
        if text.strip():
            word_before_cursor = text.strip().split()[-1]
        return word_before_cursor

    def _prioritize_required_args(self, completions):
        required_args = self._get_required_args(completions)
        optional_args = self._get_optional_args(completions)
        return required_args + optional_args

    def _get_required_args(self, completions):
        results = [
            arg for arg in completions
            if arg.required
        ]
        return results

    def _get_optional_args(self, completions):
        results = [
            arg for arg in completions
            if not arg.required
        ]
        return results

    def _get_display_text(self, completion):
        if completion.display_text is not None:
            return completion.display_text
        display_text = completion.name
        if completion.name.startswith('--') and completion.required:
            display_text += ' (required)'
        return display_text

    def _get_display_meta(self, completion):
        display_meta = ''
        type_name = getattr(completion, 'cli_type_name', None)
        help_text = getattr(completion, 'help_text', None)
        if type_name:
            display_meta += f'[{type_name}] '
        if help_text:
            display_meta += f'{help_text}'
        return display_meta

    def _filter_completions(self, completions):
        completions = self._filter_out_autoprompt_overrides(completions)
        completions = self._remove_duplicate_completions(completions)
        return completions

    def _filter_out_autoprompt_overrides(self, completions):
        filtered_completions = [
            completion for completion in completions
            if completion.name not in ['--cli-auto-prompt',
                                       '--no-cli-auto-prompt']
        ]
        return filtered_completions

    def _remove_duplicate_completions(self, completions):
        unique_completion_names = []
        unique_completions = []
        for completion in completions:
            if completion.name not in unique_completion_names:
                unique_completion_names.append(completion.name)
                unique_completions.append(completion)
        return unique_completions

    def _get_starting_location_of_last_word(self, text_before_cursor,
                                            word_before_cursor):
        if text_before_cursor and text_before_cursor[-1] == ' ':
            location = 0
        else:
            location = -len(word_before_cursor)
        return location



# Copyright 2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

import mock

from tests.unit.customizations.emr import EMRBaseAWSCommandParamsTest as \
    BaseAWSCommandParamsTest
from copy import deepcopy


class TestCreateHBaseBackup(BaseAWSCommandParamsTest):
    prefix = 'emr create-hbase-backup'
    steps = [{
        'HadoopJarStep': {
            'Args': [
                'emr.hbase.backup.Main',
                '--backup',
                '--backup-dir',
                's3://abc/'
            ],
            'Jar': '/home/hadoop/lib/hbase.jar'
        },
        'Name': 'Backup HBase',
        'ActionOnFailure': 'CANCEL_AND_WAIT'
    }]

    def test_create_hbase_backup(self):
        args = ' --cluster-id j-ABCD --dir s3://abc/'
        cmdline = self.prefix + args
        result = {'JobFlowId': 'j-ABCD', 'Steps': self.steps}

        self.assert_params_for_cmd(cmdline, result)

    def test_create_hbase_backup_consitent(self):
        args = ' --cluster-id j-ABCD --dir s3://abc/ --consistent'
        cmdline = self.prefix + args

        steps = deepcopy(self.steps)
        steps[0]['HadoopJarStep']['Args'].append('--consistent')
        result = {'JobFlowId': 'j-ABCD', 'Steps': steps}

        self.assert_params_for_cmd(cmdline, result)

    @mock.patch('awscli.customizations.emr.'
                'emrutils.get_release_label')
    def test_unsupported_command_on_release_based_cluster_error(
            self, grl_patch):
        grl_patch.return_value = 'emr-4.0'
        args = ' --cluster-id j-ABCD --dir s3://abc/'
        cmdline = self.prefix + args
        expected_error_msg = ("\naws: error: create-hbase-backup"
                              " is not supported with 'emr-4.0' release.\n")
        result = self.run_cmd(cmdline, 252)

        self.assertEqual(result[1], expected_error_msg)

if __name__ == "__main__":
    unittest.main()
